/**
 * The MeCab PHP extension
 *
 * Copyright (c) 2006-2007 Ryusuke SEKIYAMA. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any personobtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * @package     php_mecab
 * @author      Ryusuke SEKIYAMA <rsky0711@gmail.com>
 * @copyright   2006-2007 Ryusuke SEKIYAMA
 * @license     http://www.opensource.org/licenses/mit-license.php  MIT License
 * @version     SVN: $Id:$
 */

#define PHP_MECAB_MODULE_VERSION "0.2.0"

#include "php_mecab.h"
#include "php_mecab_compat.h"

#ifndef PHP_MECAB_VERSION_STRING
#define PHP_MECAB_VERSION_STRING "unknown"
#endif

#define php_mecab_error(errcode, ...) php_error_docref(NULL TSRMLS_CC, errcode, __VA_ARGS__)

#ifdef PHP_MECAB_DEBUG
#define php_mecab_c "php_mecab/mecab.c"
#define __debug_print_current_func(func, line) \
	fprintf(stderr, "%s: DEBUG: Enter function %s (%d)\n", php_mecab_c, (func), (line))
#else
#define __debug_print_current_func(func, line)
#endif

/* {{{ resource entries */

static int le_mecab;
static int le_mecab_node;
static int le_mecab_path;

/* }}} */

#ifdef ZEND_ENGINE_2
/* {{{ class entries */

extern ZEND_API zend_class_entry *zend_ce_aggregate;
extern ZEND_API zend_class_entry *zend_ce_iterator;
extern PHPAPI zend_class_entry *spl_ce_BadMethodCallException;
extern PHPAPI zend_class_entry *spl_ce_InvalidArgumentException;
extern PHPAPI zend_class_entry *spl_ce_OutOfRangeException;
extern PHPAPI zend_class_entry *reflection_function_ptr;
extern PHPAPI zend_class_entry *reflection_method_ptr;

static zend_class_entry *mecab_ce = NULL; /* alias of MeCab_Tagger */
static zend_class_entry *mecab_tagger_ce = NULL;
static zend_class_entry *mecab_node_ce = NULL;
static zend_class_entry *mecab_iterator_ce = NULL;
static zend_class_entry *mecab_path_ce = NULL;

static zend_object_handlers _php_mecab_object_handlers;
static zend_object_handlers _php_mecab_node_object_handlers;
static zend_object_handlers _php_mecab_path_object_handlers;

/* }}} */
#endif

/* {{{ type definitions */

typedef struct _php_mecab {
	mecab_t *ptr;
	char *str;
	int len;
	int ref;
} php_mecab;

typedef struct _php_mecab_node {
	php_mecab *tagger;
	const mecab_node_t *ptr;
} php_mecab_node;

typedef struct _php_mecab_path {
	php_mecab *tagger;
	const mecab_path_t *ptr;
} php_mecab_path;

typedef enum {
	ELEM_ALL,
	ELEM_SURFACE,
	ELEM_FEATURE,
	ELEM_ID,
	ELEM_LENGTH,
	ELEM_RLENGTH,
	ELEM_RCATTR,
	ELEM_LCATTR,
	ELEM_POSID,
	ELEM_CHAR_TYPE,
	ELEM_STAT,
	/*ELEM_ISBEST,
	ELEM_SENTENCE_LENGTH,
	ELEM_ALPHA,
	ELEM_BETA,
	ELEM_PROB,*/
	ELEM_WCOST,
	ELEM_COST
} php_mecab_element_type;

typedef enum {
	NODE_PREV,
	NODE_NEXT,
	NODE_ENEXT,
	NODE_BNEXT,
	NODE_RPATH,
	NODE_LPATH
} php_mecab_node_rel;

typedef enum {
	PATH_RNODE,
	PATH_RNEXT,
	PATH_LNODE,
	PATH_LNEXT
} php_mecab_path_rel;

#ifdef ZEND_ENGINE_2

typedef enum {
	TRAVERSE_NEXT,
	TRAVERSE_ENEXT,
	TRAVERSE_BNEXT
} php_mecab_traverse_mode;

typedef struct _php_mecab_object {
	zend_object std;
	php_mecab *ptr;
} php_mecab_object;

typedef struct _php_mecab_node_object {
	zend_object std;
	php_mecab_node *ptr;
	const mecab_node_t *root;
	php_mecab_traverse_mode mode;
} php_mecab_node_object;

typedef struct _php_mecab_path_object {
	zend_object std;
	php_mecab_path *ptr;
} php_mecab_path_object;

#endif

/* }}} typedefs */

/* {{{ internal function prototypes */

/* allocate for mecab */
static php_mecab *
_php_mecab_ctor(TSRMLS_D);

/* free the mecab */
static void
_php_mecab_dtor(php_mecab *mecab TSRMLS_DC);

/* free the mecab resource */
static void
_php_mecab_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC);

/* set string to the mecab */
static void
_php_mecab_set_string(php_mecab *mecab, const char *str, int len TSRMLS_DC);

/* allocate for mecab_node */
static php_mecab_node *
_php_mecab_node_ctor(TSRMLS_D);

/* free the mecab_node */
static void
_php_mecab_node_dtor(php_mecab_node *node TSRMLS_DC);

/* free the mecab_node resource */
static void
_php_mecab_node_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC);

/* set mecab to the mecab_node */
static void
_php_mecab_node_set_tagger(php_mecab_node *node, php_mecab *mecab TSRMLS_DC);

/* allocate for mecab_path */
static php_mecab_path *
_php_mecab_path_ctor(TSRMLS_D);

/* free the mecab_path */
static void
_php_mecab_path_dtor(php_mecab_path *path TSRMLS_DC);

/* free the mecab_path resource */
static void
_php_mecab_path_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC);

/* set mecab_node to the mecab_path */
static void
_php_mecab_path_set_tagger(php_mecab_path *path, php_mecab *mecab TSRMLS_DC);

/* get sibling node from mecab_node */
static zval *
_php_mecab_node_get_sibling(zval *zv, zval *object, php_mecab_node *xnode,
		php_mecab_node_rel rel TSRMLS_DC);

/* get related path from mecab_node */
static zval *
_php_mecab_node_get_path(zval *zv, zval *object, php_mecab_node *xnode,
		php_mecab_node_rel rel TSRMLS_DC);

/* get sibling path from mecab_path */
static zval *
_php_mecab_path_get_sibling(zval *zv, zval *object, php_mecab_path *xpath,
		php_mecab_path_rel rel TSRMLS_DC);

/* get related node from mecab_path */
static zval *
_php_mecab_path_get_node(zval *zv, zval *object, php_mecab_path *xpath,
		php_mecab_path_rel rel TSRMLS_DC);

/* wrappers */
static void _php_mecab_node_get_sibling_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_node_rel rel);
static void _php_mecab_node_get_path_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_node_rel rel);
static void _php_mecab_path_get_sibling_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_path_rel rel);
static void _php_mecab_path_get_node_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_path_rel rel);

/* get begin/end node list resource from mecab_node */
static void
_php_mecab_node_list_func(INTERNAL_FUNCTION_PARAMETERS, zend_bool end);

#ifdef ZEND_ENGINE_2

/* get begin/end node list object from mecab_node */
static void
_php_mecab_node_list_method(INTERNAL_FUNCTION_PARAMETERS, zend_bool end);

/* allocate for mecab object */
zend_object_value
_php_mecab_object_new(zend_class_entry *ce TSRMLS_DC);

/* free the mecab object */
static void
_php_mecab_free_object_storage(void *object TSRMLS_DC);

/* allocate for mecab_node object */
zend_object_value
_php_mecab_node_object_new(zend_class_entry *ce TSRMLS_DC);

/* free the mecab_node object */
static void
_php_mecab_node_free_object_storage(void *object TSRMLS_DC);

/* allocate for mecab_path object */
zend_object_value
_php_mecab_path_object_new(zend_class_entry *ce TSRMLS_DC);

/* free the mecab_path object */
static void
_php_mecab_path_free_object_storage(void *object TSRMLS_DC);

/* setup the object */
static void
_php_mecab_object_init(zval *zv, zend_class_entry *ce TSRMLS_DC);

/* get the node element type from annoation */
static php_mecab_element_type
_php_mecab_get_required_element_type(zval *filter TSRMLS_CC);

#endif /* ZEND_ENGINE_2 */

/* check file/dicectory accessibility */
static inline int
_php_mecab_check_path(const char *path, char *realpath_buf TSRMLS_DC);

#define php_mecab_ctor() _php_mecab_ctor(TSRMLS_C)
#define php_mecab_dtor(mecab) _php_mecab_dtor(mecab TSRMLS_CC)
#define php_mecab_set_string(mecab, str, len) _php_mecab_set_string(mecab, str, len TSRMLS_CC)
#define php_mecab_node_ctor() _php_mecab_node_ctor(TSRMLS_C)
#define php_mecab_node_dtor(node) _php_mecab_node_dtor(node TSRMLS_CC)
#define php_mecab_node_set_tagger(node, mecab) _php_mecab_node_set_tagger(node, mecab TSRMLS_CC)
#define php_mecab_path_ctor() _php_mecab_path_ctor(TSRMLS_C)
#define php_mecab_path_dtor(path) _php_mecab_path_dtor(path TSRMLS_CC)
#define php_mecab_path_set_tagger(path, mecab) _php_mecab_path_set_tagger(path, mecab TSRMLS_CC)
#define php_mecab_node_get_sibling(zv, obj, xnode, rel) _php_mecab_node_get_sibling(zv, obj, xnode, rel TSRMLS_CC)
#define php_mecab_node_get_path(zv, obj, xnode, rel) _php_mecab_node_get_path(zv, obj, xnode, rel TSRMLS_CC)
#define php_mecab_path_get_sibling(zv, obj, xpath, rel) _php_mecab_path_get_sibling(zv, obj, xpath, rel TSRMLS_CC)
#define php_mecab_path_get_node(zv, obj, xpath, rel) _php_mecab_path_get_node(zv, obj, xpath, rel TSRMLS_CC)
#define php_mecab_node_get_sibling_wrapper(rel) _php_mecab_node_get_sibling_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU, rel)
#define php_mecab_node_get_path_wrapper(rel) _php_mecab_node_get_path_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU, rel)
#define php_mecab_path_get_sibling_wrapper(rel) _php_mecab_path_get_sibling_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU, rel)
#define php_mecab_path_get_node_wrapper(rel) _php_mecab_path_get_node_wrapper(INTERNAL_FUNCTION_PARAM_PASSTHRU, rel)
#define php_mecab_node_list_func(end) _php_mecab_node_list_func(INTERNAL_FUNCTION_PARAM_PASSTHRU, end)
#ifdef ZEND_ENGINE_2
#define php_mecab_node_list_method(end) _php_mecab_node_list_method(INTERNAL_FUNCTION_PARAM_PASSTHRU, end)
#define php_mecab_object_init(zv, ce) _php_mecab_object_init(zv, ce TSRMLS_CC)
#define php_mecab_get_required_element_type(filter) _php_mecab_get_required_element_type(filter TSRMLS_CC)
#endif
#define php_mecab_check_path(path, realpath_buf) _php_mecab_check_path(path, realpath_buf TSRMLS_CC)

/* }}} prototypes */

/* {{{ argument informations */
#ifdef ZEND_BEGIN_ARG_INFO

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab__mecab, 0)
	ZEND_ARG_INFO(0, mecab)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_split, 0, 0, 1)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, dicdir)
	ZEND_ARG_INFO(0, userdic)
	ZEND_ARG_INFO(0, filter)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_new, 0, 0, 0)
	ZEND_ARG_ARRAY_INFO(0, arg, 0)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_sparse_tostr, 0, 0, 2)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_sparse_tostr_m, 0, 0, 1)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_sparse_tonode, 0, 0, 2)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_sparse_tonode_m, 0, 0, 1)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_sparse_tostr, 0, 0, 3)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, n)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_sparse_tostr_m, 0, 0, 2)
	ZEND_ARG_INFO(0, n)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_init, 0, 0, 2)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_init_m, 0, 0, 1)
	ZEND_ARG_INFO(0, str)
	ZEND_ARG_INFO(0, len)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_next_tostr, 0, 0, 1)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_nbest_next_tostr_m, 0, 0, 0)
	ZEND_ARG_INFO(0, olen)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_format_node, 0)
	ZEND_ARG_INFO(0, mecab)
	ZEND_ARG_INFO(0, node)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_format_node_m, 0)
	ZEND_ARG_OBJ_INFO(0, node, MeCab_Node, 0)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_node__node, 0)
	ZEND_ARG_INFO(0, node)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_node_toarray, 0, 0, 1)
	ZEND_ARG_INFO(0, node)
	ZEND_ARG_INFO(0, dump_all)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_mecab_node_toarray_m, 0, 0, 0)
	ZEND_ARG_INFO(0, dump_all)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_node__list, 0)
	ZEND_ARG_INFO(0, node)
	ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_node__list_m, 0)
	ZEND_ARG_INFO(0, index)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_node_settraverse, 0)
	ZEND_ARG_INFO(0, traverse)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab_path__path, 0)
	ZEND_ARG_INFO(0, path)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_mecab__magic_getter, 0)
	ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO()

#else /* ZEND_BEGIN_ARG_INFO */

#define arginfo_mecab__mecab NULL
#define arginfo_mecab_split NULL
#define arginfo_mecab_new NULL
#define arginfo_mecab_sparse_tostr NULL
#define arginfo_mecab_sparse_tonode NULL
#define arginfo_mecab_nbest_sparse_tostr NULL
#define arginfo_mecab_nbest_init NULL
#define arginfo_mecab_nbest_next_tostr NULL
#define arginfo_mecab_format_node NULL
#define arginfo_mecab_node__node NULL
#define arginfo_mecab_node_toarray NULL
#define arginfo_mecab_node__list NULL
#define arginfo_mecab_path__path NULL

#endif /* ZEND_BEGIN_ARG_INFO */
/* }}} arginfo */

#ifdef ZEND_ENGINE_2

/* redefine PHP_ME_MAPPING macro */
#if (PHP_MAJOR_VERSION == 5) && (PHP_MINOR_VERSION < 2)
#undef PHP_ME_MAPPING
#define PHP_ME_MAPPING(name, func_name, arg_types, flags) \
	ZEND_FENTRY(name, ZEND_FN(func_name), arg_types, flags)
#endif

/* {{{ MeCab methods[] */
static zend_function_entry mecab_methods[] = {
	/* MeCab API wrappers */
	PHP_ME_MAPPING(version,         mecab_version,              NULL,                               ZEND_ACC_PUBLIC | ZEND_ACC_STATIC)
	PHP_ME_MAPPING(split,           mecab_split,                arginfo_mecab_split,                ZEND_ACC_PUBLIC | ZEND_ACC_STATIC)
	PHP_ME_MAPPING(__construct,     mecab_new,                  arginfo_mecab_new,                  ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME_MAPPING(parse,           mecab_sparse_tostr,         arginfo_mecab_sparse_tostr_m,       ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(parseToString,   mecab_sparse_tostr,         arginfo_mecab_sparse_tostr_m,       ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(parseToNode,     mecab_sparse_tonode,        arginfo_mecab_sparse_tonode_m,      ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(parseNBest,      mecab_nbest_sparse_tostr,   arginfo_mecab_nbest_sparse_tostr_m, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(parseNBestInit,  mecab_nbest_init,           arginfo_mecab_nbest_init_m,         ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(next,            mecab_nbest_next_tostr,     arginfo_mecab_nbest_next_tostr_m,   ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(nextNode,        mecab_nbest_next_tonode,    NULL,                               ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(formatNode,      mecab_format_node,          arginfo_mecab_format_node_m,        ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(dictionaryInfo,  mecab_dictionary_info,      NULL,                               ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
/* }}} */

/* {{{ MeCab_Node methods[] */
static zend_function_entry mecab_node_methods[] = {
	/* Constructor */
	PHP_ME(MeCab_Node, __construct, NULL, ZEND_ACC_PRIVATE | ZEND_ACC_CTOR)
	/* Overloading implementations */
	PHP_ME(MeCab_Node, __get,   arginfo_mecab__magic_getter, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_Node, __isset, arginfo_mecab__magic_getter, ZEND_ACC_PUBLIC)
	/* IteratorAggregate implementations */
	PHP_ME(MeCab_Node, getIterator,                   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_Node, setTraverse, arginfo_mecab_node_settraverse, ZEND_ACC_PUBLIC)
	/* Dumper */
	PHP_ME_MAPPING(toArray,     mecab_node_toarray,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(toString,    mecab_node_tostring,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(__toString,  mecab_node_tostring,  NULL, ZEND_ACC_PUBLIC)
	/* Getters */
	PHP_ME_MAPPING(getPrev,     mecab_node_prev,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getNext,     mecab_node_next,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getENext,    mecab_node_enext,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getBNext,    mecab_node_bnext,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getRPath,    mecab_node_rpath,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getLPath,    mecab_node_lpath,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_Node,  getBeginNodeList,   arginfo_mecab_node__list_m, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_Node,  getEndNodeList,     arginfo_mecab_node__list_m, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getSurface,  mecab_node_surface,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getFeature,  mecab_node_feature,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getId,       mecab_node_id,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getLength,   mecab_node_length,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getRLength,  mecab_node_rlength,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getRcAttr,   mecab_node_rcattr,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getLcAttr,   mecab_node_lcattr,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getPosId,    mecab_node_posid,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getCharType, mecab_node_char_type, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getStat,     mecab_node_stat,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(isBest,      mecab_node_isbest,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getSentenceLength,   mecab_node_sentence_length, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getAlpha,    mecab_node_alpha,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getBeta,     mecab_node_beta,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getProb,     mecab_node_prob,      NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getWCost,    mecab_node_wcost,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getCost,     mecab_node_cost,      NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
/* }}} */

/* {{{ MeCab_NodeIterator methods[] */
static zend_function_entry mecab_iterator_methods[] = {
	/* Constructor */
	PHP_ME(MeCab_NodeIterator, __construct, NULL, ZEND_ACC_PRIVATE | ZEND_ACC_CTOR)
	/* Iterator implementations */
	PHP_ME(MeCab_NodeIterator,  current,    NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_NodeIterator,  key,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_NodeIterator,  next,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_NodeIterator,  rewind,     NULL, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_NodeIterator,  valid,      NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
/* }}} */

/* {{{ MeCab_Path methods[] */
static zend_function_entry mecab_path_methods[] = {
	/* Constructor */
	PHP_ME(MeCab_Path, __construct, NULL, ZEND_ACC_PRIVATE | ZEND_ACC_CTOR)
	/* Overloading implementations */
	PHP_ME(MeCab_Path, __get,   arginfo_mecab__magic_getter, ZEND_ACC_PUBLIC)
	PHP_ME(MeCab_Path, __isset, arginfo_mecab__magic_getter, ZEND_ACC_PUBLIC)
	/* Getters */
	PHP_ME_MAPPING(getRNext, mecab_path_rnext, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getLNext, mecab_path_lnext, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getRNode, mecab_path_rnode, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getLNode, mecab_path_lnode, NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getProb,  mecab_path_prob,  NULL, ZEND_ACC_PUBLIC)
	PHP_ME_MAPPING(getCost,  mecab_path_cost,  NULL, ZEND_ACC_PUBLIC)
	{ NULL, NULL, NULL }
};
/* }}} methods */

#endif /* ZEND_ENGINE_2 */
/* }}} class definitions */

/* {{{ mecab_functions[] */
function_entry mecab_functions[] = {
	/* Get MeCab library version */
	PHP_FE(mecab_version,               NULL)
	/* Wakati-Gaki function */
	PHP_FE(mecab_split,                 arginfo_mecab_split)
	/* MeCab API wrappers */
	PHP_FE(mecab_new,                   arginfo_mecab_new)
	PHP_FE(mecab_destroy,               arginfo_mecab__mecab)
	PHP_FE(mecab_sparse_tostr,          arginfo_mecab_sparse_tostr)
	PHP_FE(mecab_sparse_tonode,         arginfo_mecab_sparse_tonode)
	PHP_FE(mecab_nbest_sparse_tostr,    arginfo_mecab_nbest_sparse_tostr)
	PHP_FE(mecab_nbest_init,            arginfo_mecab_nbest_init)
	PHP_FE(mecab_nbest_next_tostr,      arginfo_mecab_nbest_next_tostr)
	PHP_FE(mecab_nbest_next_tonode,     arginfo_mecab__mecab)
	PHP_FE(mecab_format_node,           arginfo_mecab_format_node)
	PHP_FE(mecab_dictionary_info,       arginfo_mecab__mecab)
	/* Dumper for mecab_node */
	PHP_FE(mecab_node_toarray,          arginfo_mecab_node_toarray)
	PHP_FE(mecab_node_tostring,         arginfo_mecab_node__node)
	/* Getters for mecab_node */
	PHP_FE(mecab_node_prev,             arginfo_mecab_node__node)
	PHP_FE(mecab_node_next,             arginfo_mecab_node__node)
	PHP_FE(mecab_node_enext,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_bnext,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_rpath,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_lpath,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_begin_node_list,  arginfo_mecab_node__list)
	PHP_FE(mecab_node_end_node_list,    arginfo_mecab_node__list)
	PHP_FE(mecab_node_surface,          arginfo_mecab_node__node)
	PHP_FE(mecab_node_feature,          arginfo_mecab_node__node)
	PHP_FE(mecab_node_id,               arginfo_mecab_node__node)
	PHP_FE(mecab_node_length,           arginfo_mecab_node__node)
	PHP_FE(mecab_node_rlength,          arginfo_mecab_node__node)
	PHP_FE(mecab_node_rcattr,           arginfo_mecab_node__node)
	PHP_FE(mecab_node_lcattr,           arginfo_mecab_node__node)
	PHP_FE(mecab_node_posid,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_char_type,        arginfo_mecab_node__node)
	PHP_FE(mecab_node_stat,             arginfo_mecab_node__node)
	PHP_FE(mecab_node_isbest,           arginfo_mecab_node__node)
	PHP_FE(mecab_node_sentence_length,  arginfo_mecab_node__node)
	PHP_FE(mecab_node_alpha,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_beta,             arginfo_mecab_node__node)
	PHP_FE(mecab_node_prob,             arginfo_mecab_node__node)
	PHP_FE(mecab_node_wcost,            arginfo_mecab_node__node)
	PHP_FE(mecab_node_cost,             arginfo_mecab_node__node)
	/* Aliases */
	/* Getters for mecab_path */
	PHP_FE(mecab_path_rnext,            arginfo_mecab_path__path)
	PHP_FE(mecab_path_lnext,            arginfo_mecab_path__path)
	PHP_FE(mecab_path_rnode,            arginfo_mecab_path__path)
	PHP_FE(mecab_path_lnode,            arginfo_mecab_path__path)
	PHP_FE(mecab_path_prob,             arginfo_mecab_path__path)
	PHP_FE(mecab_path_cost,             arginfo_mecab_path__path)
	{ NULL, NULL, NULL }
};
/* }}} */

/* {{{ cross-extension dependencies */

#if ZEND_EXTENSION_API_NO >= 220050617
static zend_module_dep mecab_deps[] = {
	ZEND_MOD_REQUIRED("reflection")
	ZEND_MOD_REQUIRED("spl")
	{NULL, NULL, NULL, 0}
};
#endif
/* }}} */

/* {{{ mecab_module_entry */
zend_module_entry mecab_module_entry = {
#if ZEND_EXTENSION_API_NO >= 220050617
	STANDARD_MODULE_HEADER_EX, NULL,
	mecab_deps,
#else
	STANDARD_MODULE_HEADER,
#endif
	"mecab",
	mecab_functions,
	PHP_MINIT(mecab),
	PHP_MSHUTDOWN(mecab),
	NULL,
	NULL,
	PHP_MINFO(mecab),
	PHP_MECAB_MODULE_VERSION,
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_MECAB
ZEND_GET_MODULE(mecab)
#endif

/* {{{ globals and ini entries */
ZEND_DECLARE_MODULE_GLOBALS(mecab)

PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY("mecab.default_rcfile", "", PHP_INI_ALL,
		OnUpdateString, default_rcfile, zend_mecab_globals, mecab_globals)
	STD_PHP_INI_ENTRY("mecab.default_dicdir", "", PHP_INI_ALL,
		OnUpdateString, default_dicdir, zend_mecab_globals, mecab_globals)
	STD_PHP_INI_ENTRY("mecab.default_userdic", "", PHP_INI_ALL,
		OnUpdateString, default_userdic, zend_mecab_globals, mecab_globals)
PHP_INI_END()

static void
_php_mecab_init_globals(zend_mecab_globals *mecab_globals)
{
	mecab_globals->default_rcfile = NULL;
	mecab_globals->default_dicdir = NULL;
	mecab_globals->default_userdic = NULL;
}

static void
_php_mecab_shutdown_globals(zend_mecab_globals *mecab_globals)
{
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION */
PHP_MINIT_FUNCTION(mecab)
{
	ZEND_INIT_MODULE_GLOBALS(mecab, _php_mecab_init_globals, _php_mecab_shutdown_globals)

	REGISTER_INI_ENTRIES();

	REGISTER_STRING_CONSTANT("MECAB_VERSION", (char *)mecab_version(), CONST_PERSISTENT | CONST_CS);
#define PHP_MECAB_REGISTER_CONSTANT(name) \
	REGISTER_LONG_CONSTANT(#name, name, CONST_PERSISTENT | CONST_CS)
	PHP_MECAB_REGISTER_CONSTANT(MECAB_NOR_NODE);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_UNK_NODE);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_BOS_NODE);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_EOS_NODE);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_SYS_DIC);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_USR_DIC);
	PHP_MECAB_REGISTER_CONSTANT(MECAB_UNK_DIC);
#undef PHP_MECAB_REGISTER_CONSTANT

	le_mecab = zend_register_list_destructors_ex(
			_php_mecab_free_resource, NULL, "mecab", module_number);
	le_mecab_node = zend_register_list_destructors_ex(
			_php_mecab_node_free_resource, NULL, "mecab_node", module_number);
	le_mecab_path = zend_register_list_destructors_ex(
			_php_mecab_path_free_resource, NULL, "mecab_path", module_number);

#ifdef ZEND_ENGINE_2
	{
		zend_class_entry ce1, ce1a;

		INIT_CLASS_ENTRY(ce1a, "MeCab", mecab_methods);
		ce1a.create_object = _php_mecab_object_new;
		mecab_ce = zend_register_internal_class(&ce1a);

		INIT_CLASS_ENTRY(ce1, "MeCab_Tagger", mecab_methods);
		ce1.create_object = _php_mecab_object_new;
		mecab_tagger_ce = zend_register_internal_class(&ce1);

		memcpy(&_php_mecab_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
		_php_mecab_object_handlers.clone_obj = NULL;

		zend_declare_class_constant_string(mecab_ce, "VERSION", 7, (char *)mecab_version() TSRMLS_DC);
		zend_declare_class_constant_long(mecab_ce, "SYS_DIC", 7, MECAB_SYS_DIC TSRMLS_CC);
		zend_declare_class_constant_long(mecab_ce, "USR_DIC", 7, MECAB_USR_DIC TSRMLS_CC);
		zend_declare_class_constant_long(mecab_ce, "UNK_DIC", 7, MECAB_UNK_DIC TSRMLS_CC);

		zend_declare_class_constant_string(mecab_tagger_ce, "VERSION", 7, (char *)mecab_version() TSRMLS_DC);
		zend_declare_class_constant_long(mecab_tagger_ce, "SYS_DIC", 7, MECAB_SYS_DIC TSRMLS_CC);
		zend_declare_class_constant_long(mecab_tagger_ce, "USR_DIC", 7, MECAB_USR_DIC TSRMLS_CC);
		zend_declare_class_constant_long(mecab_tagger_ce, "UNK_DIC", 7, MECAB_UNK_DIC TSRMLS_CC);
	}
	{
		zend_class_entry ce2, ce2i;

		INIT_CLASS_ENTRY(ce2, "MeCab_Node", mecab_node_methods);
		ce2.create_object = _php_mecab_node_object_new;
		mecab_node_ce = zend_register_internal_class(&ce2);

		INIT_CLASS_ENTRY(ce2i, "MeCab_NodeIterator", mecab_iterator_methods);
		ce2i.create_object = _php_mecab_node_object_new;
		mecab_iterator_ce = zend_register_internal_class(&ce2i);

		memcpy(&_php_mecab_node_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
		_php_mecab_node_object_handlers.clone_obj = NULL;

		zend_class_implements(mecab_node_ce TSRMLS_CC, 1, zend_ce_aggregate);
		zend_class_implements(mecab_iterator_ce TSRMLS_CC, 1, zend_ce_iterator);

		zend_declare_class_constant_long(mecab_node_ce, "NOR", 3, MECAB_NOR_NODE TSRMLS_CC);
		zend_declare_class_constant_long(mecab_node_ce, "UNK", 3, MECAB_UNK_NODE TSRMLS_CC);
		zend_declare_class_constant_long(mecab_node_ce, "BOS", 3, MECAB_BOS_NODE TSRMLS_CC);
		zend_declare_class_constant_long(mecab_node_ce, "EOS", 3, MECAB_EOS_NODE TSRMLS_CC);

		zend_declare_class_constant_long(mecab_node_ce, "TRAVERSE_NEXT", 13, (long)TRAVERSE_NEXT TSRMLS_CC);
		zend_declare_class_constant_long(mecab_node_ce, "TRAVERSE_ENEXT", 14, (long)TRAVERSE_ENEXT TSRMLS_CC);
		zend_declare_class_constant_long(mecab_node_ce, "TRAVERSE_BNEXT", 14, (long)TRAVERSE_BNEXT TSRMLS_CC);
	}
	{
		zend_class_entry ce3;

		INIT_CLASS_ENTRY(ce3, "MeCab_Path", mecab_path_methods);
		ce3.create_object = _php_mecab_path_object_new;
		mecab_path_ce = zend_register_internal_class(&ce3);

		memcpy(&_php_mecab_path_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
		_php_mecab_path_object_handlers.clone_obj = NULL;
	}
#endif

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION */
PHP_MSHUTDOWN_FUNCTION(mecab)
{
	UNREGISTER_INI_ENTRIES();
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION */
PHP_MINFO_FUNCTION(mecab)
{
	php_info_print_table_start();
	php_info_print_table_row(2, "MeCab Support", "enabled");
	php_info_print_table_row(2, "Module Version", PHP_MECAB_MODULE_VERSION);
	php_info_print_table_end();

	php_info_print_table_start();
	php_info_print_table_header(3, "Version Info", "Compiled", "Linked");
	php_info_print_table_row(3, "MeCab Library", PHP_MECAB_VERSION_STRING, mecab_version());
	php_info_print_table_end();

	DISPLAY_INI_ENTRIES();
}
/* }}} */

/* {{{ internal function implementation for mecab_t */

/* {{{ _php_mecab_ctor()
 * allocate for mecab
 */
static php_mecab *
_php_mecab_ctor(TSRMLS_D)
{
	php_mecab *mecab = NULL;

	mecab = (php_mecab *)ecalloc(1, sizeof(php_mecab));
	if (mecab == NULL) {
		return NULL;
	}

	mecab->ptr = NULL;
	mecab->str = NULL;
	mecab->len = 0;
	mecab->ref = 1;

	return mecab;
}
/* }}} */

/* {{{ _php_mecab_dtor()
 * free the mecab
 */
static void
_php_mecab_dtor(php_mecab *mecab TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
#ifdef PHP_MECAB_DEBUG
	fprintf(stderr, "%s: DEBUG: Reducing refcount: %d->%d\n",
			php_mecab_c, mecab->ref, mecab->ref - 1);
#endif
	mecab->ref--;
	if (mecab->ref == 0) {
		if (mecab->str != NULL) {
#ifdef PHP_MECAB_DEBUG
			fprintf(stderr, "%s: DEBUG: Free string buffer %p (%p)\n",
					php_mecab_c, mecab->str, &(mecab->str));
#endif
			efree(mecab->str);
		}
#ifdef PHP_MECAB_DEBUG
		fprintf(stderr, "%s: DEBUG: Destroy mecab object %p (%p)\n",
				php_mecab_c, mecab->ptr, &(mecab->ptr));
#endif
		mecab_destroy(mecab->ptr);
		efree(mecab);
	}
}
/* }}} */

/* {{{ php_mecab_free_resource()
 * free the mecab resource
 */
static void
_php_mecab_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_dtor((php_mecab *)rsrc->ptr);
}
/* }}} */

/* {{{ _php_mecab_set_string()
 * set string to the mecab
 */
static void
_php_mecab_set_string(php_mecab *mecab, const char *str, int len TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (mecab->str != NULL) {
		efree(mecab->str);
	}
	if (str == NULL) {
		mecab->str = NULL;
		mecab->len = 0;
	} else {
		mecab->str = estrndup(str, len);
		mecab->len = (size_t)len;
	}
}
/* }}} */

#ifdef ZEND_ENGINE_2

/* {{{ _php_mecab_object_new()
 * allocate for mecab object
 */
zend_object_value
_php_mecab_object_new(zend_class_entry *ce TSRMLS_DC)
{
	php_mecab_object *intern = NULL;
	zend_object_value retval;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	intern = (php_mecab_object *)ecalloc(1, sizeof(php_mecab_object));
	intern->ptr = php_mecab_ctor();

	zend_object_std_init(&intern->std, ce TSRMLS_CC);
	zend_hash_copy(intern->std.properties, &ce->default_properties,
			(copy_ctor_func_t)zval_add_ref, (void *)NULL, sizeof(zval *));

	retval.handle = zend_objects_store_put((void *)intern,
			(zend_objects_store_dtor_t)zend_objects_destroy_object,
			(zend_objects_free_object_storage_t)_php_mecab_free_object_storage,
			NULL TSRMLS_CC);
	retval.handlers = &_php_mecab_object_handlers;

	return retval;
}
/* }}} */

/* {{{ _php_mecab_free_object_storage()
 * free the mecab object
 */
static void
_php_mecab_free_object_storage(void *object TSRMLS_DC)
{
	php_mecab_object *intern = (php_mecab_object *)object;
	__debug_print_current_func(__FUNCTION__, __LINE__);
	zend_object_std_dtor(&intern->std TSRMLS_CC);
	php_mecab_dtor(intern->ptr);
	efree(object);
}
/* }}} */

#endif /* ZEND_ENGINE_2 */
/* }}} mecab_t */

/* {{{ internal function implementation for mecab_node_t */

/* {{{ _php_mecab_node_ctor()
 * allocate for mecab_node
 */
static php_mecab_node *
_php_mecab_node_ctor(TSRMLS_D)
{
	php_mecab_node *node = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	node = (php_mecab_node *)ecalloc(1, sizeof(php_mecab_node));
	if (node == NULL) {
		return NULL;
	}

	node->tagger = NULL;
	node->ptr = NULL;

	return node;
}
/* }}} */

/* {{{ _php_mecab_node_dtor()
 * free the mecab_node
 */
static void
_php_mecab_node_dtor(php_mecab_node *node TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (node->tagger != NULL) {
		php_mecab_dtor(node->tagger);
	}
	efree(node);
}
/* }}} */

/* {{{ _php_mecab_node_free_resource()
 * free the mecab_node resource
 */
static void
_php_mecab_node_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	_php_mecab_node_dtor((php_mecab_node *)rsrc->ptr TSRMLS_CC);
}
/* }}} */

/* {{{ _php_mecab_node_set_tagger()
 * set mecab to the mecab_node
 */
static void
_php_mecab_node_set_tagger(php_mecab_node *node, php_mecab *mecab TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
#ifdef PHP_MECAB_DEBUG
	fprintf(stderr, "%s: DEBUG: Incrementing refcount on %s: %d->%d\n",
			php_mecab_c, __FUNCTION__, mecab->ref, mecab->ref + 1);
#endif
	if (node->tagger != NULL) {
		php_mecab_dtor(node->tagger);
	}
	if (mecab == NULL) {
		node->tagger = NULL;
	} else {
		node->tagger = mecab;
		node->tagger->ref++;
	}
}
/* }}} */

#ifdef ZEND_ENGINE_2

/* {{{ _php_mecab_node_object_new()
 * allocate for mecab_node object
 */
zend_object_value
_php_mecab_node_object_new(zend_class_entry *ce TSRMLS_DC)
{
	php_mecab_node_object *intern = NULL;
	zend_object_value retval;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	intern = (php_mecab_node_object *)ecalloc(1, sizeof(php_mecab_node_object));
	intern->ptr = php_mecab_node_ctor();
	intern->root = NULL;
	intern->mode = TRAVERSE_NEXT;

	zend_object_std_init(&intern->std, ce TSRMLS_CC);
	zend_hash_copy(intern->std.properties, &ce->default_properties,
			(copy_ctor_func_t)zval_add_ref, (void *)NULL, sizeof(zval *));

	retval.handle = zend_objects_store_put((void *)intern,
			(zend_objects_store_dtor_t)zend_objects_destroy_object,
			(zend_objects_free_object_storage_t)_php_mecab_node_free_object_storage,
			NULL TSRMLS_CC);
	retval.handlers = &_php_mecab_node_object_handlers;

	return retval;
}
/* }}} */

/* {{{ _php_mecab_node_free_object_storage()
 * free the mecab_node object
 */
static void
_php_mecab_node_free_object_storage(void *object TSRMLS_DC)
{
	php_mecab_node_object *intern = (php_mecab_node_object *)object;
	__debug_print_current_func(__FUNCTION__, __LINE__);
	zend_object_std_dtor(&intern->std TSRMLS_CC);
	php_mecab_node_dtor(intern->ptr);
	efree(object);
}
/* }}} */

#endif /* ZEND_ENGINE_2 */
/* }}} mecab_node_t */

/* {{{ _php_mecab_path_ctor()
 * allocate for mecab_path
 */
static php_mecab_path *
_php_mecab_path_ctor(TSRMLS_D)
{
	php_mecab_path *path = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	path = (php_mecab_path *)ecalloc(1, sizeof(php_mecab_path));
	if (path == NULL) {
		return NULL;
	}

	path->tagger = NULL;
	path->ptr = NULL;

	return path;
}
/* }}} */

/* {{{ _php_mecab_path_dtor()
 * free the mecab_path
 */
static void
_php_mecab_path_dtor(php_mecab_path *path TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (path->tagger != NULL) {
		php_mecab_dtor(path->tagger);
	}
	efree(path);
}
/* }}} */

/* {{{ _php_mecab_path_free_resource()
 * free the mecab_path resource
 */
static void
_php_mecab_path_free_resource(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	_php_mecab_path_dtor((php_mecab_path *)rsrc->ptr TSRMLS_CC);
}
/* }}} */

/* {{{ _php_mecab_path_set_tagger()
 * set mecab_node to the mecab_path
 */
static void
_php_mecab_path_set_tagger(php_mecab_path *path, php_mecab *mecab TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
#ifdef PHP_MECAB_DEBUG
	fprintf(stderr, "%s: DEBUG: Incrementing refcount on %s: %d->%d\n",
			php_mecab_c, __FUNCTION__, mecab->ref, mecab->ref + 1);
#endif
	if (path->tagger != NULL) {
		php_mecab_dtor(path->tagger);
	}
	if (mecab == NULL) {
		path->tagger = NULL;
	} else {
		path->tagger = mecab;
		path->tagger->ref++;
	}
}
/* }}} */

#ifdef ZEND_ENGINE_2

/* {{{ _php_mecab_path_object_new()
 * allocate for mecab_path object
 */
zend_object_value
_php_mecab_path_object_new(zend_class_entry *ce TSRMLS_DC)
{
	php_mecab_path_object *intern = NULL;
	zend_object_value retval;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	intern = (php_mecab_path_object *)ecalloc(1, sizeof(php_mecab_path_object));
	intern->ptr = php_mecab_path_ctor();

	zend_object_std_init(&intern->std, ce TSRMLS_CC);
	zend_hash_copy(intern->std.properties, &ce->default_properties,
			(copy_ctor_func_t)zval_add_ref, (void *)NULL, sizeof(zval *));

	retval.handle = zend_objects_store_put((void *)intern,
			(zend_objects_store_dtor_t)zend_objects_destroy_object,
			(zend_objects_free_object_storage_t)_php_mecab_path_free_object_storage,
			NULL TSRMLS_CC);
	retval.handlers = &_php_mecab_path_object_handlers;

	return retval;
}
/* }}} */

/* {{{ _php_mecab_path_free_object_storage()
 * free the mecab_path object
 */
static void
_php_mecab_path_free_object_storage(void *object TSRMLS_DC)
{
	php_mecab_path_object *intern = (php_mecab_path_object *)object;
	__debug_print_current_func(__FUNCTION__, __LINE__);
	zend_object_std_dtor(&intern->std TSRMLS_CC);
	php_mecab_path_dtor(intern->ptr);
	efree(object);
}
/* }}} */

/* {{{ _php_mecab_object_init()
 * setup the object
 */
static void
_php_mecab_object_init(zval *zv, zend_class_entry *ce TSRMLS_DC)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	zval_dtor(zv);
	Z_TYPE_P(zv) = IS_OBJECT;
	object_init_ex(zv, ce);
	zv->refcount = 1;
	zv->is_ref = 1;
}

/* }}} */

#endif /* ZEND_ENGINE_2 */
/* }}} mecab_path_t */

/* {{{ _php_mecab_node_get_sibling()
 * get sibling node from mecab_node
 */
static zval *
_php_mecab_node_get_sibling(zval *zv, zval *object, php_mecab_node *xnode,
		php_mecab_node_rel rel TSRMLS_DC)
{
	const mecab_node_t *node = xnode->ptr;
	php_mecab_node *xsbl = NULL;
	const mecab_node_t *sbl = NULL;
	zval *retval = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (zv == NULL) {
		MAKE_STD_ZVAL(retval);
	} else {
		zval_dtor(zv);
		retval = zv;
	}

	/* scan */
	if (rel == NODE_PREV) {
		sbl = node->prev;
	} else if (rel == NODE_NEXT) {
		sbl = node->next;
	} else if (rel == NODE_ENEXT) {
		sbl = node->enext;
	} else if (rel == NODE_BNEXT) {
		sbl = node->bnext;
	} else {
		ZVAL_FALSE(retval);
		return;
	}

	if (sbl == NULL) {
		ZVAL_NULL(retval);
		return retval;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_node_object *newobj;
		php_mecab_object_init(retval, mecab_node_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, retval);
		xsbl = newobj->ptr;
#endif
	} else {
		xsbl = php_mecab_node_ctor();
		ZEND_REGISTER_RESOURCE(retval, xsbl, le_mecab_node);
	}
	xsbl->ptr = sbl;
	php_mecab_node_set_tagger(xsbl, xnode->tagger);

	return retval;
}
/* }}} */

/* {{{ _php_mecab_node_get_path()
 * get related path from mecab_node
 */
static zval *
_php_mecab_node_get_path(zval *zv, zval *object, php_mecab_node *xnode,
		php_mecab_node_rel rel TSRMLS_DC)
{
	const mecab_node_t *node = xnode->ptr;
	php_mecab_path *xpath = NULL;
	const mecab_path_t *path = NULL;
	zval *retval = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (zv == NULL) {
		MAKE_STD_ZVAL(retval);
	} else {
		zval_dtor(zv);
		retval = zv;
	}

	/* scan */
	if (rel == NODE_RPATH) {
		path = node->rpath;
	} else if (rel == NODE_LPATH) {
		path = node->lpath;
	} else {
		ZVAL_FALSE(retval);
		return retval;
	}

	if (path == NULL) {
		ZVAL_NULL(retval);
		return retval;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_path_object *newobj;
		php_mecab_object_init(retval, mecab_path_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_path_object *, retval);
		xpath = newobj->ptr;
#endif
	} else {
		xpath = php_mecab_path_ctor();
		ZEND_REGISTER_RESOURCE(retval, xpath, le_mecab_path);
	}
	xpath->ptr = path;
	php_mecab_path_set_tagger(xpath, xnode->tagger);

	return retval;
}
/* }}} */

/* {{{ _php_mecab_path_get_sibling()
 * get sibling path from mecab_path
 */
static zval *
_php_mecab_path_get_sibling(zval *zv, zval *object, php_mecab_path *xpath,
		php_mecab_path_rel rel TSRMLS_DC)
{
	const mecab_path_t *path = xpath->ptr;
	php_mecab_path *xsbl = NULL;
	const mecab_path_t *sbl = NULL;
	zval *retval = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (zv == NULL) {
		MAKE_STD_ZVAL(retval);
	} else {
		zval_dtor(zv);
		retval = zv;
	}

	/* scan */
	if (rel == PATH_RNEXT) {
		sbl = path->rnext;
	} else if (rel == PATH_LNEXT) {
		sbl = path->lnext;
	} else {
		ZVAL_FALSE(retval);
		return retval;
	}

	if (sbl == NULL) {
		ZVAL_NULL(retval);
		return retval;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_path_object *newobj;
		php_mecab_object_init(retval, mecab_path_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_path_object *, retval);
		xsbl = newobj->ptr;
#endif
	} else {
		xsbl = php_mecab_path_ctor();
		ZEND_REGISTER_RESOURCE(retval, xsbl, le_mecab_path);
	}
	xsbl->ptr = sbl;
	php_mecab_path_set_tagger(xsbl, xpath->tagger);

	return retval;
}
/* }}} */

/* {{{ _php_mecab_path_get_node()
 * get related node from mecab_path
 */
static zval *
_php_mecab_path_get_node(zval *zv, zval *object, php_mecab_path *xpath,
		php_mecab_path_rel rel TSRMLS_DC)
{
	const mecab_path_t *path = xpath->ptr;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;
	zval *retval = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (zv == NULL) {
		MAKE_STD_ZVAL(retval);
	} else {
		zval_dtor(zv);
		retval = zv;
	}

	/* scan */
	if (rel == PATH_RNODE) {
		node = path->rnode;
	} else if (rel == PATH_LNODE) {
		node = path->lnode;
	} else {
		ZVAL_FALSE(retval);
		return retval;
	}

	if (node == NULL) {
		ZVAL_NULL(retval);
		return retval;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_node_object *newobj;
		php_mecab_object_init(retval, mecab_node_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, retval);
		xnode = newobj->ptr;
#endif
	} else {
		xnode = php_mecab_node_ctor();
		ZEND_REGISTER_RESOURCE(retval, xnode, le_mecab_node);
	}
	xnode->ptr = node;
	php_mecab_node_set_tagger(xnode, xpath->tagger);

	return retval;
}
/* }}} */

/* {{{ _php_mecab_node_get_sibling_wrapper()
 * wraps php_mecab_node_get_sibling()
 */
static void
_php_mecab_node_get_sibling_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_node_rel rel)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_NODE_FROM_PARAMETER();

	php_mecab_node_get_sibling(return_value, object, xnode, rel);
}
/* }}} */

/* {{{ _php_mecab_node_get_path_wrapper()
 * wraps php_mecab_node_get_path()
 */
static void
_php_mecab_node_get_path_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_node_rel rel)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_NODE_FROM_PARAMETER();

	php_mecab_node_get_path(return_value, object, xnode, rel);
}
/* }}} */

/* {{{ _php_mecab_path_get_sibling_wrapper()
 * wraps php_mecab_path_get_sibling()
 */
static void
_php_mecab_path_get_sibling_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_path_rel rel)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zpath = NULL;
	php_mecab_path *xpath = NULL;
	const mecab_path_t *path = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PATH_FROM_PARAMETER();

	php_mecab_path_get_sibling(return_value, object, xpath, rel);
}
/* }}} */

/* {{{ _php_mecab_path_get_node_wrapper()
 * wraps php_mecab_path_get_path()
 */
static void
_php_mecab_path_get_node_wrapper(INTERNAL_FUNCTION_PARAMETERS, php_mecab_path_rel rel)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zpath = NULL;
	php_mecab_path *xpath = NULL;
	const mecab_path_t *path = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PATH_FROM_PARAMETER();

	php_mecab_path_get_node(return_value, object, xpath, rel);
}
/* }}} */

/* {{{ _php_mecab_node_list_func()
 * get begin/end node list resource from mecab_node
 */
static void
_php_mecab_node_list_func(INTERNAL_FUNCTION_PARAMETERS, zend_bool end)
{
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;
	long index = 0;

	php_mecab_node *newnode = NULL;
	const mecab_node_t *newptr = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl",
			&znode, &index) == FAILURE)
	{
		return;
	}
	ZEND_FETCH_RESOURCE(xnode, php_mecab_node *, &znode, -1, "mecab_node", le_mecab_node);
	node = xnode->ptr;

	if (node->stat != MECAB_BOS_NODE) {
		php_mecab_error(E_WARNING, "Only available in BOS node");
		return;
	}

	if (index < 0 || index > (long)node->sentence_length) {
		php_mecab_error(E_WARNING, "Index should be in range from 0 to %u: %ld given",
				node->sentence_length, index);
		return;
	}

	newptr = (end) ? node->end_node_list[index] : node->begin_node_list[index];
	if (newptr == NULL) {
		RETURN_NULL();
	}

	newnode = php_mecab_node_ctor();
	newnode->ptr = newptr;
	php_mecab_node_set_tagger(newnode, xnode->tagger);
	ZEND_REGISTER_RESOURCE(return_value, newnode, le_mecab_node);
}

#ifdef ZEND_ENGINE_2
/* {{{ _php_mecab_node_list_method()
 * get begin/end node list object from mecab_node
 */
static void
_php_mecab_node_list_method(INTERNAL_FUNCTION_PARAMETERS, zend_bool end)
{
	php_mecab_node_object *intern = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;
	long index = 0;

	php_mecab_node_object *newobj = NULL;
	php_mecab_node *newnode = NULL;
	const mecab_node_t *newptr = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_set_error_handling(EH_THROW, spl_ce_InvalidArgumentException TSRMLS_CC);
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l",
			&index) == FAILURE)
	{
		php_std_error_handling();
		return;
	}
	php_std_error_handling();

	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	if (node->stat != MECAB_BOS_NODE) {
		zend_throw_exception_ex(spl_ce_BadMethodCallException, 0 TSRMLS_CC,
				"Only available in BOS node");
		return;
	}

	if (index < 0 || index > (long)node->sentence_length) {
		zend_throw_exception_ex(spl_ce_OutOfRangeException, 0 TSRMLS_CC,
				"Index should be in range from 0 to %u: %ld given",
				node->sentence_length, index);
	}

	newptr = (end) ? node->end_node_list[index] : node->begin_node_list[index];
	if (newptr == NULL) {
		RETURN_NULL();
	}

	php_mecab_object_init(return_value, mecab_node_ce);
	newobj = (php_mecab_node_object *)zend_object_store_get_object(return_value TSRMLS_CC);
	newnode = newobj->ptr;
	newnode->ptr = newptr;
	php_mecab_node_set_tagger(newnode, xnode->tagger);
}
/* }}} */

#define ANNOTATION_FINDER "@mecab_element_require"
#define ANNOTATION_FINDER_LEN (sizeof(ANNOTATION_FINDER) - 1)

static int
_element_type_is(const char *annotation, const char *name, size_t length)
{
	if (strncmp(annotation, name, length)) {
		return 0;
	}
	annotation += length;
	if (*annotation == '\0' || *annotation == '\r' || *annotation == '\n' ||
		*annotation == '\t' || *annotation == ' ')
	{
		return 1;
	}
	return 0;
}

#define element_type_is(annotation, name) _element_type_is(annotation, #name, sizeof(#name) -1)

/* {{{ _php_mecab_get_required_element_type()
 * get the node element type from annoation
 */
static php_mecab_element_type
_php_mecab_get_required_element_type(zval *filter TSRMLS_CC)
{
	php_mecab_element_type type = ELEM_ALL;
	char *doc_comment = NULL;
	char *doc_comment_end = NULL;
	zval *reflector = NULL;
	zval *retval = NULL;
	zval *ctor = NULL;
	zval *get_doc_comment = NULL;
	zval **ctor_argv[2] = { NULL, NULL };

	/* instantiate reflector */
	MAKE_STD_ZVAL(reflector);
	Z_TYPE_P(reflector) = IS_OBJECT;
	if (Z_TYPE_P(filter) == IS_ARRAY) {
		object_init_ex(reflector, reflection_method_ptr);
	} else {
		object_init_ex(reflector, reflection_function_ptr);
	}
	reflector->refcount = 1;
	reflector->is_ref = 1;

	/* create a function name "__construct()" */
	MAKE_STD_ZVAL(ctor);
	ZVAL_STRINGL(ctor, "__construct", sizeof("__construct") - 1, 1);

	/* call __construct() */
	if (Z_TYPE_P(filter) == IS_ARRAY) {
		HashTable *class_method = Z_ARRVAL_P(filter);
		zval **entry = NULL;
		int ctor_argc = 0;

		zend_hash_internal_pointer_reset(class_method);
		while (ctor_argc < 2 && zend_hash_get_current_data(class_method, (void **)&entry) == SUCCESS) {
			ctor_argv[ctor_argc++] = entry;
			zend_hash_move_forward(class_method);
		}

		/* call ReflectionMethod::__construct() */
		if (call_user_function_ex(NULL, &reflector, ctor,
			&retval, ctor_argc, ctor_argv, 0, NULL TSRMLS_CC) == FAILURE)
		{
			zval_ptr_dtor(&reflector);
			zval_ptr_dtor(&ctor);
			return ELEM_ALL;
		}
	} else {
		ctor_argv[0] = &filter;

		/* call ReflectionFunction::__construct() */
		if (call_user_function_ex(NULL, &reflector, ctor,
			&retval, 1, ctor_argv, 0, NULL TSRMLS_CC) == FAILURE)
		{
			zval_ptr_dtor(&reflector);
			zval_ptr_dtor(&ctor);
			return ELEM_ALL;
		}
	}

	/* free */
	if (retval != NULL) {
		zval_ptr_dtor(&retval);
	}
	zval_ptr_dtor(&ctor);

	/* create a function name "getDocComment()" */
	MAKE_STD_ZVAL(get_doc_comment);
	ZVAL_STRINGL(get_doc_comment, "getDocComment", sizeof("getDocComment") - 1, 1);

	/* call ReflectionFunction::getDocComment() or ReflectionMethod::getDocComment() */
	if (call_user_function_ex(NULL, &reflector, get_doc_comment,
		&retval, 0, NULL, 0, NULL TSRMLS_CC) == FAILURE || retval == NULL)
	{
		zval_ptr_dtor(&reflector);
		zval_ptr_dtor(&get_doc_comment);
		return ELEM_ALL;
	}

	/* free */
	zval_ptr_dtor(&reflector);
	zval_ptr_dtor(&get_doc_comment);

	/* cast to string */
	convert_to_string(retval);
	doc_comment = Z_STRVAL_P(retval);
	doc_comment_end = doc_comment + Z_STRLEN_P(retval);

	/* check for the annotation finder */
	while (doc_comment < doc_comment_end - ANNOTATION_FINDER_LEN) {
		if (!strncmp(doc_comment, ANNOTATION_FINDER, ANNOTATION_FINDER_LEN)) {
			doc_comment += ANNOTATION_FINDER_LEN;
			if (*doc_comment == ' ' || *doc_comment == '\t') {
				do {
					doc_comment++;
				} while (*doc_comment == ' ' || *doc_comment == '\t');
			} else {
				doc_comment = doc_comment_end;
			}
			break;
		}
		doc_comment++;
	}

	/* check for the annotation */
	if (doc_comment == doc_comment_end) {
		type = ELEM_ALL;
	} else if (element_type_is(doc_comment, surface)) {
		type = ELEM_SURFACE;
	} else if (element_type_is(doc_comment, feature)) {
		type = ELEM_FEATURE;
	} else if (element_type_is(doc_comment, id)) {
		type = ELEM_ID;
	} else if (element_type_is(doc_comment, length)) {
		type = ELEM_LENGTH;
	} else if (element_type_is(doc_comment, rlength)) {
		type = ELEM_RLENGTH;
	} else if (element_type_is(doc_comment, lcAttr)) {
		type = ELEM_RCATTR;
	} else if (element_type_is(doc_comment, lcAttr)) {
		type = ELEM_LCATTR;
	} else if (element_type_is(doc_comment, posid)) {
		type = ELEM_POSID;
	} else if (element_type_is(doc_comment, char_type)) {
		type = ELEM_CHAR_TYPE;
	} else if (element_type_is(doc_comment, stat)) {
		type = ELEM_STAT;
	/*} else if (element_type_is(doc_comment, isbest)) {
		type = ELEM_ISBEST;
	} else if (element_type_is(doc_comment, sentence_length)) {
		type = ELEM_SENTENCE_LENGTH;
	} else if (element_type_is(doc_comment, alpha)) {
		type = ELEM_ALPHA;
	} else if (element_type_is(doc_comment, beta)) {
		type = ELEM_BETA;
	} else if (element_type_is(doc_comment, prob)) {
		type = ELEM_PROB;*/
	} else if (element_type_is(doc_comment, wcost)) {
		type = ELEM_WCOST;
	} else if (element_type_is(doc_comment, cost)) {
		type = ELEM_COST;
	} else {
		type = ELEM_ALL;
	}

	/* free */
	zval_ptr_dtor(&retval);

	return type;
}
/* }}} */
#endif

/* {{{ _php_mecab_node_list_method()
 * check file/dicectory accessibility
 */
static inline int
_php_mecab_check_path(const char *path, char *realpath_buf TSRMLS_DC)
{
	if (VCWD_ACCESS(path, F_OK) != 0 || VCWD_ACCESS(path, R_OK) != 0) {
		return FAILURE;
	}
	if (!VCWD_REALPATH(path, realpath_buf)) {
		return FAILURE;
	}
	if (php_check_open_basedir(realpath_buf TSRMLS_CC)) {
		return FAILURE;
	}
#if PHP_MAJOR_VERSION < 6
	if (PG(safe_mode) && !php_checkuid(realpath_buf, NULL, CHECKUID_CHECK_FILE_AND_DIR)) {
		return FAILURE;
	}
#endif
	return SUCCESS;
}
/* }}} */

/* {{{ macro for checking constructor options */

/* check if default parameter is set */
#define PHP_MECAB_CHECK_DEFAULT(name) \
	(MECAB_G(default_##name) != NULL && MECAB_G(default_##name)[0] != '\0')

/* check for option */
#define PHP_MECAB_CHECK_OPTION(option) \
{ \
	if (!strcmp((option), "-r") || !strcmp((option), "--rcfile")) { \
		/* resource file */ \
		resolved_path = rcfile_buf; \
		flag_expected = 0; \
		path_expected = 1; \
	} else if (!strcmp((option), "-d") || !strcmp((option), "--dicdir")) { \
		/* system dicdir */ \
		resolved_path = dicdir_buf; \
		flag_expected = 0; \
		path_expected = 1; \
	} else if (!strcmp((option), "-u") || !strcmp((option), "--userdic")) { \
		/* user dictionary */ \
		resolved_path = userdic_buf; \
		flag_expected = 0; \
		path_expected = 1; \
	} else if ( /* options whose parameter is not filename */ \
		!strcmp((option), "-l") || !strcmp((option), "--lattice-level") || \
		!strcmp((option), "-O") || !strcmp((option), "--output-format-type") || \
		!strcmp((option), "-F") || !strcmp((option), "--node-format") || \
		!strcmp((option), "-U") || !strcmp((option), "--unk-format") || \
		!strcmp((option), "-B") || !strcmp((option), "--bos-format") || \
		!strcmp((option), "-E") || !strcmp((option), "--eos-format") || \
		!strcmp((option), "-x") || !strcmp((option), "--unk-feature") || \
		!strcmp((option), "-b") || !strcmp((option), "--input-buffer-size") || \
		!strcmp((option), "-N") || !strcmp((option), "--nbest") || \
		!strcmp((option), "-t") || !strcmp((option), "--theta")) \
	{ \
		flag_expected = 0; \
		path_expected = 0; \
	} else if ( /* options which does not have parameter */ \
		!strcmp((option), "-a") || !strcmp((option), "--all-morphs") || \
		!strcmp((option), "-p") || !strcmp((option), "--partial") || \
		!strcmp((option), "-C") || !strcmp((option), "--allocate-sentence")) \
	{ \
		flag_expected = 1; \
		path_expected = 0; \
	} else { /* invalid options */ \
		efree(argv); \
		php_mecab_error(E_WARNING, "Invalid option '%s' given", (option)); \
		RETURN_FALSE; \
	} \
}

/* check for open_basedir and safe_mode */
#define PHP_MECAB_CHECK_FILE(path) \
{ \
	if (php_mecab_check_path((path), resolved_path) == FAILURE) { \
		efree(argv); \
		php_mecab_error(E_WARNING, "'%s' does not exist or is not readable", (path)); \
		RETURN_FALSE; \
	} \
	flag_expected = 1; \
	path_expected = 0; \
}

/* }}} */

/* {{{ Functions */

/* {{{ proto string mecab_version(void) */
/**
 * string mecab_version(void)
 *
 * Get the version number of MeCab.
 *
 * @return	string	The version of linked MeCab library.
 */
PHP_FUNCTION(mecab_version)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	RETURN_STRING((char *)mecab_version(), 1);
}
/* }}} mecab_version */

/* {{{ proto array mecab_split(string str[, string dicdir[, string userdic[, callback filter]]]) */
/**
 * array mecab_split(string str[, string dicdir[, string userdic[, callback filter]]])
 * array MeCab_Tagger::split(string str[, string dicdir[, string userdic[, callback filter]]])
 *
 * Split string into an array of morphemes.
 *
 * @param	string	$str	The target string.
 * @param	string	$dicdir	The path for system dictionary.
 * @param	string	$userdic	The path for user dictionary.
 * @param	callback	$filter	A filter function or method.
 * @return	array
 */
PHP_FUNCTION(mecab_split)
{
	/* variables from argument */
	const char *str = NULL;
	int str_len = 0;
	const char *dicdir = NULL;
	int dicdir_len = 0;
	const char *userdic = NULL;
	int userdic_len = 0;
	zval *filter = NULL;

	/* local variables */
	mecab_t *mecab = NULL;
	const mecab_node_t *node = NULL;
	int argc = 2;
	char *argv[5] = { "mecab", "-Owakati", NULL, NULL, NULL };
	char pathbuf[2][MAXPATHLEN + 3] = { '\0' };
	char *dicdir_buf = &(pathbuf[0][0]);
	char *userdic_buf = &(pathbuf[1][0]);
	char *filter_name = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|ssz/",
			&str, &str_len, &dicdir, &dicdir_len, &userdic, &userdic_len, &filter) == FAILURE)
	{
		return;
	}

	/* check for callback */
	if (filter != NULL) {
		if (!zend_is_callable(filter, 0, &filter_name)) {
			if (filter_name != NULL) {
				php_mecab_error(E_WARNING, "Argument #4 '%s' is not a valid callback", filter_name);
				efree(filter_name);
			} else {
				php_mecab_error(E_WARNING, "Argument #4 is not a valid callback");
			}
			return;
		}
		if (filter_name != NULL) {
			efree(filter_name);
		}
	}

	/* apply default options */
	if (dicdir_len == 0 && PHP_MECAB_CHECK_DEFAULT(dicdir)) {
		dicdir = MECAB_G(default_dicdir);
		dicdir_len = (int)strlen(MECAB_G(default_dicdir));
	}
	if (dicdir_len == 0 && PHP_MECAB_CHECK_DEFAULT(userdic)) {
		userdic = MECAB_G(default_userdic);
		userdic_len = (int)strlen(MECAB_G(default_userdic));
	}

	/* check for dictionary */
	if (dicdir != NULL && dicdir_len > 0) {
		char *dicdir_ptr = dicdir_buf;
		*dicdir_ptr++ = '-';
		*dicdir_ptr++ = 'd';
		if (php_mecab_check_path(dicdir, dicdir_ptr) == FAILURE) {
			php_mecab_error(E_WARNING, "'%s' does not exist or is not readable", dicdir);
			RETURN_FALSE;
		}
		argv[argc++] = dicdir_buf;
	}
	if (userdic != NULL && userdic_len > 0) {
		char *userdic_ptr = userdic_buf;
		*userdic_ptr++ = '-';
		*userdic_ptr++ = 'u';
		if (php_mecab_check_path(userdic, userdic_ptr) == FAILURE) {
			php_mecab_error(E_WARNING, "'%s' does not exist or is not readable", userdic);
			RETURN_FALSE;
		}
		argv[argc++] = userdic_buf;
	}

	/* create mecab object */
	mecab = mecab_new(argc, argv);

#ifdef PHP_MECAB_DEBUG
	{
		char **argp = argv;
		fprintf(stderr, "%s: DEBUG: argc = %d\n", php_mecab_c, argc);
		while (*argp) {
			fprintf(stderr, "%s: DEBUG: argv[%d] = %s\n", php_mecab_c, (int)(argp - argv), *argp);
			argp++;
		}
	}
#endif

	/* on error */
	if (mecab == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(NULL));
		RETURN_FALSE;
	}

	/* parse the string */
	node = mecab_sparse_tonode(mecab, str);
	if (node == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		mecab_destroy(mecab);
		RETURN_FALSE;
	}

	/* initialize */
	array_init(return_value);

	/* put surfaces of each node to return value */
	if (filter == NULL) {
		while (node != NULL) {
			if (node->length > 0) {
				add_next_index_stringl(return_value, (char *)node->surface, (int)node->length, 1);
			}
			node = node->next;
		}
	} else {
		while (node != NULL) {
			zval **argv[] = { NULL };
			zval *arg0 = NULL;
			zval *retval = NULL;
#ifdef ZEND_ENGINE_2
			php_mecab_element_type type = 0;
#endif

			MAKE_STD_ZVAL(arg0);

#ifdef ZEND_ENGINE_2
			type = php_mecab_get_required_element_type(filter);

			switch (type) {
			  case ELEM_SURFACE:
				ZVAL_STRINGL(arg0, (char *)node->surface, (int)node->length, 1);
				break;
			  case ELEM_FEATURE:
				ZVAL_STRING(arg0, (char *)node->feature, 1);
				break;
			  case ELEM_ID:
				ZVAL_LONG(arg0, (long)node->id);
				break;
			  case ELEM_LENGTH:
				ZVAL_LONG(arg0, (long)node->length);
				break;
			  case ELEM_RLENGTH:
				ZVAL_LONG(arg0, (long)node->rlength);
				break;
			  case ELEM_RCATTR:
				ZVAL_LONG(arg0, (long)node->rcAttr);
				break;
			  case ELEM_LCATTR:
				ZVAL_LONG(arg0, (long)node->lcAttr);
				break;
			  case ELEM_POSID:
				ZVAL_LONG(arg0, (long)node->posid);
				break;
			  case ELEM_CHAR_TYPE:
				ZVAL_LONG(arg0, (long)node->char_type);
				break;
			  case ELEM_STAT:
				ZVAL_LONG(arg0, (long)node->stat);
				break;
			/*  case ELEM_ISBEST:
				ZVAL_LONG(arg0, (long)node->isbest);
				break;
			  case ELEM_SENTENCE_LENGTH:
				if (node->stat == MECAB_BOS_NODE) {
					ZVAL_LONG(arg0, (long)node->sentence_length);
				} else {
					ZVAL_NULL(arg0);
				}
				break;
			  case ELEM_ALPHA:
				ZVAL_DOUBLE(arg0, (double)node->alpha);
				break;
			  case ELEM_BETA:
				ZVAL_DOUBLE(arg0, (double)node->beta);
				break;
			  case ELEM_PROB:
				ZVAL_DOUBLE(arg0, (double)node->prob);
				break;*/
			  case ELEM_WCOST:
				ZVAL_LONG(arg0, (long)node->wcost);
				break;
			  case ELEM_COST:
				ZVAL_LONG(arg0, (long)node->cost);
				break;
			  case ELEM_ALL:
			  default:
#endif
				MAKE_STD_ZVAL(arg0);
				array_init(arg0);

				add_assoc_stringl(arg0, "surface",   (char *)node->surface, (int)node->length, 1);
				add_assoc_string(arg0,  "feature",   (char *)node->feature, 1);
				add_assoc_long(arg0,    "id",        (long)node->id);
				add_assoc_long(arg0,    "length",    (long)node->length);
				add_assoc_long(arg0,    "rlength",   (long)node->rlength);
				add_assoc_long(arg0,    "rcAttr",    (long)node->rcAttr);
				add_assoc_long(arg0,    "lcAttr",    (long)node->lcAttr);
				add_assoc_long(arg0,    "posid",     (long)node->posid);
				add_assoc_long(arg0,    "char_type", (long)node->char_type);
				add_assoc_long(arg0,    "stat",      (long)node->stat);
				/*add_assoc_bool(arg0,    "isbest",    (long)node->isbest);
				if (node->stat == MECAB_BOS_NODE) {
					add_assoc_long(arg0, "sentence_length", (long)node->sentence_length);
				} else {
					add_assoc_null(arg0, "sentence_length");
				}
				add_assoc_double(arg0,  "alpha",     (double)node->alpha);
				add_assoc_double(arg0,  "beta",      (double)node->beta);
				add_assoc_double(arg0,  "prob",      (double)node->prob);*/
				add_assoc_long(arg0,    "wcost",     (long)node->wcost);
				add_assoc_long(arg0,    "cost",      (long)node->cost);
#ifdef ZEND_ENGINE_2
			}
#endif

			argv[0] = &arg0;

			if (call_user_function_ex(EG(function_table), NULL, filter,
					&retval, 1, argv, 0, NULL TSRMLS_CC) == SUCCESS && retval != NULL)
			{
				convert_to_boolean(retval);
				if (Z_BVAL_P(retval) == 1 && node->length > 0) {
					add_next_index_stringl(return_value, (char *)node->surface, (int)node->length, 1);
				}
				zval_ptr_dtor(&retval);
			}

			zval_ptr_dtor(&arg0);

			node = node->next;
		}
	}

	/* free mecab object */
	mecab_destroy(mecab);
}
/* }}} mecab_split */

/* {{{ proto resource mecab mecab_new([array options]) */
/**
 * resource mecab mecab_new([array options])
 * object MeCab_Tagger MeCab_Tagger::__construct([array options])
 *
 * Create new tagger resource of MeCab.
 *
 * @param	array	$options	The analysis/output options. (optional)
 *								The values are same to command line options.
 *								The detail is found in the web site and/or the manpage of MeCab.
 * @return	resource mecab	A tagger resource of MeCab.
 */
PHP_FUNCTION(mecab_new)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	zval *zoptions = NULL;

	/* declaration of the local variables */
	size_t min_argc = 5; /* "mecab" + "-r" + "-d" + "-u" + NULL  */
	int argc = 1;
	char **argv = NULL;
	zend_bool flag_expected = 1;
	zend_bool path_expected = 0;
	char pathbuf[3][MAXPATHLEN + 3] = { '\0' };
	char *rcfile_buf = &(pathbuf[0][0]);
	char *dicdir_buf = &(pathbuf[1][0]);
	char *userdic_buf = &(pathbuf[2][0]);
	char *resolved_path = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|a/", &zoptions) == FAILURE) {
		return;
	}

	/* parse options */
	if (zoptions != NULL) {
		HashTable *options = Z_ARRVAL_P(zoptions);
		unsigned long idx = 0;
		char *key = NULL;
		zval **entry = NULL;

		zend_hash_internal_pointer_reset(options);
		argv = (char **)emalloc(sizeof(char *) * (2 * zend_hash_num_elements(options) + min_argc));

		while (zend_hash_get_current_data(options, (void **)&entry) == SUCCESS) {
			convert_to_string_ex(entry);
			switch (zend_hash_get_current_key(options, &key, &idx, 0)) {
			  case HASH_KEY_IS_STRING:
				PHP_MECAB_CHECK_OPTION(key);
				argv[argc++] = key;
				if (path_expected) {
					PHP_MECAB_CHECK_FILE(Z_STRVAL_PP(entry));
					argv[argc++] = resolved_path;
				} else {
					argv[argc++] = Z_STRVAL_PP(entry);
				}
				flag_expected = 1;
				path_expected = 0;
				break;
			  case HASH_KEY_IS_LONG:
				if (flag_expected) {
					PHP_MECAB_CHECK_OPTION(Z_STRVAL_PP(entry));
					argv[argc++] = Z_STRVAL_PP(entry);
				} else if (path_expected) {
					PHP_MECAB_CHECK_FILE(Z_STRVAL_PP(entry));
					argv[argc++] = resolved_path;
				} else {
					argv[argc++] = Z_STRVAL_PP(entry);
				}
				break;
			}
			zend_hash_move_forward(options);
		}
	} else {
		argv = (char **)emalloc(sizeof(char *) * min_argc);
	}

	/* apply default options */
	if (rcfile_buf[0] == '\0' && PHP_MECAB_CHECK_DEFAULT(rcfile)) {
		resolved_path = rcfile_buf;
		*resolved_path++ = '-';
		*resolved_path++ = 'r';
		PHP_MECAB_CHECK_FILE(MECAB_G(default_rcfile));
		argv[argc++] = rcfile_buf;
	}
	if (dicdir_buf[0] == '\0' && PHP_MECAB_CHECK_DEFAULT(dicdir)) {
		resolved_path = dicdir_buf;
		*resolved_path++ = '-';
		*resolved_path++ = 'd';
		PHP_MECAB_CHECK_FILE(MECAB_G(default_dicdir));
		argv[argc++] = dicdir_buf;
	}
	if (userdic_buf[0] == '\0' && PHP_MECAB_CHECK_DEFAULT(userdic)) {
		resolved_path = userdic_buf;
		*resolved_path++ = '-';
		*resolved_path++ = 'u';
		PHP_MECAB_CHECK_FILE(MECAB_G(default_userdic));
		argv[argc++] = userdic_buf;
	}

	/* create mecab object */
	argv[0] = "mecab";
	argv[argc] = NULL;
	mecab = mecab_new(argc, argv);

#ifdef PHP_MECAB_DEBUG
	{
		char **argp = argv;
		fprintf(stderr, "%s: DEBUG: argc = %d\n", php_mecab_c, argc);
		while (*argp) {
			fprintf(stderr, "%s: DEBUG: argv[%d] = %s\n", php_mecab_c, (int)(argp - argv), *argp);
			argp++;
		}
	}
#endif
	efree(argv);

	/* on error */
	if (mecab == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(NULL));
		RETURN_FALSE;
	}

	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_object *intern;
		PHP_MECAB_FETCH_OBJECT(intern, php_mecab_object *, object);
		xmecab = intern->ptr;
#endif
	} else {
		xmecab = php_mecab_ctor();
		ZEND_REGISTER_RESOURCE(return_value, xmecab, le_mecab);
	}
	xmecab->ptr = mecab;
}
/* }}} mecab_new */

/* {{{ proto void mecab_destroy(resource mecab mecab) */
/**
 * void mecab mecab_destroy(resource mecab mecab)
 *
 * Free the tagger.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @return	void
 */
PHP_FUNCTION(mecab_destroy)
{
	/* declaration of the arguments */
	zval *zmecab = NULL;
	php_mecab *mecab = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zmecab) == FAILURE) {
		return;
	}
	ZEND_FETCH_RESOURCE(mecab, php_mecab *, &zmecab, -1, "mecab", le_mecab);

	/* free the resource */
	FREE_RESOURCE(zmecab);
}
/* }}} mecab_destroy */

/* {{{ proto string mecab_sparse_tostr(resource mecab mecab, string str[, int len[, int olen]]) */
/**
 * string mecab_sparse_tostr(resource mecab mecab, string str[, int len[, int olen]])
 * string MeCab_Tagger::parse(string str[, int len[, int olen]])
 * string MeCab_Tagger::parseToString(string str[, int len[, int olen]])
 *
 * Get the parse result as a string.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	string	$str	The parse target.
 * @param	int	$len	The maximum length that can be analyzed. (optional)
 * @param	int	$olen	The limit length of the output buffer. (optional)
 * @return	string	The parse result.
 *					If output buffer has overflowed, returns false.
 */
PHP_FUNCTION(mecab_sparse_tostr)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	const char *str = NULL;
	int str_len = 0;
	long len = 0;
	long olen = 0;

	/* declaration of the local variables */
	size_t ilen = 0;
	char *ostr = NULL;
	zend_bool ostr_free = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PARSE_PARAMETERS("s|ll", &str, &str_len, &len, &olen);

	/* call mecab_sparse_tostr() */
	php_mecab_set_string(xmecab, str, str_len);
	ilen = (size_t)((len > 0) ? MIN(len, (long)str_len) : str_len);
	if (olen == 0) {
		ostr = (char *)mecab_sparse_tostr2(mecab, xmecab->str, ilen);
	} else {
		ostr = (char *)emalloc((size_t)olen + 1);
		ostr = mecab_sparse_tostr3(mecab, xmecab->str, ilen, ostr, (size_t)olen);
		ostr_free = 1;
	}

	/* set return value */
	if (ostr == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		ZVAL_BOOL(return_value, 0);
	} else {
		ZVAL_STRING(return_value, ostr, 1);
	}

	/* free */
	if (ostr_free) {
		efree(ostr);
	}
}
/* }}} mecab_sparse_tostr */

/* {{{ proto resource mecab_node mecab_sparse_tonode(resource mecab mecab, string str[, int len]) */
/**
 * resource mecab_node mecab_sparse_tonode(resource mecab mecab, string str[, int len])
 * object MeCab_Node MeCab_Tagger::parseToNode(string str[, int len])
 *
 * Get the parse result as a node.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	string	$str	The parse target.
 * @param	int	$len	The maximum length that can be analyzed. (optional)
 * @return	resource mecab_node	The result node of given string.
 */
PHP_FUNCTION(mecab_sparse_tonode)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	const char *str = NULL;
	int str_len = 0;
	long len = 0;

	/* declaration of the local variables */
	size_t ilen = 0;
	const mecab_node_t *node = NULL;
	php_mecab_node *xnode = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PARSE_PARAMETERS("s|l", &str, &str_len, &len);

	/* call mecab_sparse_tonode() */
	php_mecab_set_string(xmecab, str, str_len);
	ilen = (size_t)((len > 0) ? MIN(len, (long)str_len) : str_len);
	node = mecab_sparse_tonode2(mecab, xmecab->str, ilen);
	if (node == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		RETURN_FALSE;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_node_object *newobj;
		php_mecab_object_init(return_value, mecab_node_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, return_value);
		xnode = newobj->ptr;
#endif
	} else {
		xnode = php_mecab_node_ctor();
		ZEND_REGISTER_RESOURCE(return_value, xnode, le_mecab_node);
	}
	xnode->ptr = node;
	php_mecab_node_set_tagger(xnode, xmecab);
}
/* }}} mecab_sparse_tonode */

/* {{{ proto string mecab_nbest_sparse_tostr(resource mecab mecab, int n, string str[, int len[, int olen]]) */
/**
 * string mecab_nbest_sparse_tostr(resource mecab mecab, int n, string str[, int len[, int olen]])
 * string MeCab_Tagger::parseNBest(int n, string str[, int len[, int olen]])
 *
 * Get the N-Best list as a string.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	int	$n	The number of the result list.
 * @param	string	$str	The parse target.
 * @param	int	$len	The maximum length that can be analyzed. (optional)
 * @param	int	$olen	The maximum length of the output. (optional)
 * @return	string	The N-Best list.
 *					If output buffer has overflowed, returns false.
 */
PHP_FUNCTION(mecab_nbest_sparse_tostr)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	long n = 0;
	const char *str = NULL;
	int str_len = 0;
	long len = 0;
	long olen = 0;

	/* declaration of the local variables */
	size_t ilen = 0;
	char *ostr = NULL;
	zend_bool ostr_free = 1;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PARSE_PARAMETERS("ls|ll", &n, &str, &str_len, &len, &olen);

	/* call mecab_nbest_sparse_tostr() */
	php_mecab_set_string(xmecab, str, str_len);
	ilen = (size_t)((len > 0) ? MIN(len, (long)str_len) : str_len);
	if (olen == 0) {
		ostr = (char *)mecab_nbest_sparse_tostr2(mecab, n, xmecab->str, ilen);
	} else {
		ostr = (char *)emalloc(olen + 1);
		ostr = mecab_nbest_sparse_tostr3(mecab, n, xmecab->str, ilen, ostr, (size_t)olen);
		ostr_free = 1;
	}

	/* set return value */
	if (ostr == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		ZVAL_BOOL(return_value, 0);
	} else {
		ZVAL_STRING(return_value, ostr, 1);
	}

	/* free */
	if (ostr_free) {
		efree(ostr);
	}
}
/* }}} mecab_nbest_sparse_tostr */

/* {{{ proto bool mecab_nbest_init(resource mecab mecab, string str[, int len]) */
/**
 * bool mecab_nbest_init(resource mecab mecab, string str[, int len])
 * bool MeCab_Tagger::parseNBestInit(string str[, int len])
 *
 * Initialize the N-Best list.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	string	$str	The parse target.
 * @param	int	$len	The maximum length that can be analyzed. (optional)
 * @return	bool	True if succeeded to initilalize, otherwise returns false.
 */
PHP_FUNCTION(mecab_nbest_init)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	const char *str = NULL;
	int str_len = 0;
	long len = 0;

	/* declaration of the local variables */
	size_t ilen = 0;
	int result = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PARSE_PARAMETERS("s|l", &str, &str_len, &len);

	/* call mecab_nbest_init() */
	php_mecab_set_string(xmecab, str, str_len);
	ilen = (size_t)((len > 0) ? MIN(len, (long)str_len) : str_len);
	result = mecab_nbest_init2(mecab, xmecab->str, ilen);
	if (result == 0) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		RETURN_FALSE;
	}
	RETURN_TRUE;
}
/* }}} mecab_nbest_init */

/* {{{ proto string mecab_nbest_next_tostr(resource mecab mecab[, int olen]) */
/**
 * string mecab_nbest_next_tostr(resource mecab mecab[, int olen])
 * string MeCab_Tagger::next([int olen]])
 *
 * Get the next result of N-Best as a string.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	int	$olen	The maximum length of the output. (optional)
 * @return	string	The parse result of the next pointer.
 *					If there are no more results, returns false.
 *					Also returns false if output buffer has overflowed.
 */
PHP_FUNCTION(mecab_nbest_next_tostr)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the arguments */
	long olen = 0;

	/* declaration of the local variables */
	char *ostr = NULL;
	zend_bool ostr_free = 0;
	const char *what = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_PARSE_PARAMETERS("|l", &olen);

	/* call mecab_nbest_sparse_tostr() */
	if (olen == 0) {
		ostr = (char *)mecab_nbest_next_tostr(mecab);
	} else {
		ostr = (char *)emalloc(olen + 1);
		ostr = mecab_nbest_next_tostr2(mecab, ostr, (size_t)olen);
		ostr_free = 1;
	}

	/* set return value */
	if (ostr == NULL) {
		if ((what = mecab_strerror(mecab)) != NULL &&
			strcmp((char *)what, "no more results"))
		{
			php_mecab_error(E_WARNING, "%s", what);
		}
		ZVAL_BOOL(return_value, 0);
	} else {
		ZVAL_STRING(return_value, ostr, 1);
	}

	/* free */
	if (ostr_free) {
		efree(ostr);
	}
}
/* }}} mecab_nbest_next_tostr */

/* {{{ proto resource mecab_node mecab_nbest_next_tonode(resource mecab mecab) */
/**
 * resource mecab_node mecab_nbest_next_tonode(resource mecab mecab)
 * object MeCab_Node MeCab_Tagger::nextNode(void)
 *
 * Get the next result of N-Best as a node.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @return	resource mecab_node	The result node of the next pointer.
 *								If there are no more results, returns false.
 */
PHP_FUNCTION(mecab_nbest_next_tonode)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the local variables */
	const mecab_node_t *node = NULL;
	php_mecab_node *xnode = NULL;
	const char *what = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_FROM_PARAMETER();

	/* call mecab_nbest_next_tonode() */
	node = mecab_nbest_next_tonode(mecab);
	if (node == NULL) {
		if ((what = mecab_strerror(mecab)) != NULL &&
			strcmp((char *)what, "no more results"))
		{
			php_mecab_error(E_WARNING, "%s", what);
		}
		RETURN_FALSE;
	}

	/* set return value */
	if (object) {
#ifdef ZEND_ENGINE_2
		php_mecab_node_object *newobj;
		php_mecab_object_init(return_value, mecab_node_ce);
		PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, return_value);
		xnode = newobj->ptr;
#endif
	} else {
		xnode = php_mecab_node_ctor();
		ZEND_REGISTER_RESOURCE(return_value, xnode, le_mecab_node);
	}
	xnode->ptr = node;
	php_mecab_node_set_tagger(xnode, xmecab);
}
/* }}} mecab_nbest_next_tonode */

/* {{{ proto string mecab_format_node(resource mecab mecab, resource mecab_node node) */
/**
 * string mecab_format_node(resource mecab mecab, resource mecab_node node)
 * string MeCab_Tagger::formatNode(object MeCab_Node node)
 *
 * Format a node to string.
 * The format is specified by "-O" option or --{node|unk|bos|eos}-format=STR.
 * The detail is found in the web site and/or the manpage of MeCab.
 * NOTICE: If the option was "wakati" or "dump", the return string will be empty.
 *
 * @param	resource mecab	$mecab	The tagger resource of MeCab.
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	string	The formatted string.
 * @see	mecab_node_tostring
 */
PHP_FUNCTION(mecab_format_node)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *node_object = NULL;
	zval *zmecab = NULL;
	zval *znode = NULL;
	php_mecab *xmecab = NULL;
	php_mecab_node *xnode = NULL;
	mecab_t *mecab = NULL;
	const mecab_node_t *node = NULL;

	/* declaration of the local variables */
	const char *fmt = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (object) {
#ifdef ZEND_ENGINE_2
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O",
				&node_object, mecab_node_ce) == FAILURE)
		{
			return;
		} else {
			php_mecab_object *intern;
			php_mecab_node_object *intern_node;
			PHP_MECAB_FETCH_OBJECT(intern, php_mecab_object *, object);
			PHP_MECAB_FETCH_OBJECT(intern_node, php_mecab_node_object *, node_object);
			xmecab = intern->ptr;
			xnode = intern_node->ptr;
		}
#endif
	} else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr",
				&zmecab, &znode) == FAILURE)
		{
			return;
		}
		ZEND_FETCH_RESOURCE(xmecab, php_mecab *, &zmecab, -1, "mecab", le_mecab);
		ZEND_FETCH_RESOURCE(xnode, php_mecab_node *, &znode, -1, "mecab_node", le_mecab_node);
	}
	mecab = xmecab->ptr;
	node = xnode->ptr;

	/* call mecab_format_node() */
	fmt = mecab_format_node(mecab, node);
	if (fmt == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		RETURN_FALSE;
	}

	/* set return value */
	RETURN_STRING((char *)fmt, 1);
}
/* }}} mecab_format_node */

/* {{{ proto array mecab_dictionary_info(resource mecab mecab) */
/**
 * array mecab_dictionary_info(resource mecab mecab)
 * array MeCab_Tagger::dictionaryInfo(void)
 *
 * Get the information of using dictionary as an associative array.
 *
 * @return	array	The information of the dictionary.
 */
PHP_FUNCTION(mecab_dictionary_info)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zmecab = NULL;
	php_mecab *xmecab = NULL;
	mecab_t *mecab = NULL;

	/* declaration of the local variables */
	const mecab_dictionary_info_t *dicinfo = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_FROM_PARAMETER();

	/* get dictionary information */
	dicinfo = mecab_dictionary_info(mecab);
	if (dicinfo == NULL) {
		RETURN_NULL();
	}

	/* initialize */
	array_init(return_value);

	/* get information for each dictionary */
	while (dicinfo != NULL) {
		zval *tmp = NULL;

		MAKE_STD_ZVAL(tmp);
		array_init(tmp);

		add_assoc_string(tmp, "filename", (char *)dicinfo->filename, 1);
		add_assoc_string(tmp, "charset",  (char *)dicinfo->charset, 1);
		add_assoc_long(tmp, "size",    (long)dicinfo->size);
		add_assoc_long(tmp, "type",    (long)dicinfo->type);
		add_assoc_long(tmp, "lsize",   (long)dicinfo->lsize);
		add_assoc_long(tmp, "rsize",   (long)dicinfo->rsize);
		add_assoc_long(tmp, "version", (long)dicinfo->version);

		add_next_index_zval(return_value, tmp);

		dicinfo = dicinfo->next;
	}
}
/* }}} mecab_dictionary_info */

/* {{{ proto array mecab_node_toarray(resource mecab_node node[, bool dump_all]) */
/**
 * array mecab_node_toarray(resource mecab_node node)
 * array MeCab_Node::toArray(void)
 *
 * Get all elements of the node as an associative array.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @param	bool	$dump_all	Whether dump all related nodes and paths or not.
 * @return	array	All elements of the node.
 */
PHP_FUNCTION(mecab_node_toarray)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	/* declaration of the arguments */
	zend_bool dump_all = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_NODE_PARSE_PARAMETERS("|b", &dump_all);

	/* initialize */
	array_init(return_value);

	/* assign siblings and paths */
	if (dump_all) {
		add_assoc_zval(return_value, "prev",
				php_mecab_node_get_sibling(NULL, object, xnode, NODE_PREV));
		add_assoc_zval(return_value, "next",
				php_mecab_node_get_sibling(NULL, object, xnode, NODE_NEXT));
		add_assoc_zval(return_value, "enext",
				php_mecab_node_get_sibling(NULL, object, xnode, NODE_ENEXT));
		add_assoc_zval(return_value, "bnext",
				php_mecab_node_get_sibling(NULL, object, xnode, NODE_BNEXT));
		add_assoc_zval(return_value, "rpath",
				php_mecab_node_get_path(NULL, object, xnode, NODE_RPATH));
		add_assoc_zval(return_value, "lpath",
				php_mecab_node_get_path(NULL, object, xnode, NODE_LPATH));
	}

	/* assign node info */
	add_assoc_stringl(return_value, "surface",   (char *)node->surface, (int)node->length, 1);
	add_assoc_string(return_value,  "feature",   (char *)node->feature, 1);
	add_assoc_long(return_value,    "id",        (long)node->id);
	add_assoc_long(return_value,    "length",    (long)node->length);
	add_assoc_long(return_value,    "rlength",   (long)node->rlength);
	add_assoc_long(return_value,    "rcAttr",    (long)node->rcAttr);
	add_assoc_long(return_value,    "lcAttr",    (long)node->lcAttr);
	add_assoc_long(return_value,    "posid",     (long)node->posid);
	add_assoc_long(return_value,    "char_type", (long)node->char_type);
	add_assoc_long(return_value,    "stat",      (long)node->stat);
	add_assoc_bool(return_value,    "isbest",    (long)node->isbest);
	if (node->stat == MECAB_BOS_NODE) {
		add_assoc_long(return_value, "sentence_length", (long)node->sentence_length);
	} else {
		add_assoc_null(return_value, "sentence_length");
	}
	add_assoc_double(return_value,  "alpha",     (double)node->alpha);
	add_assoc_double(return_value,  "beta",      (double)node->beta);
	add_assoc_double(return_value,  "prob",      (double)node->prob);
	add_assoc_long(return_value,    "wcost",     (long)node->wcost);
	add_assoc_long(return_value,    "cost",      (long)node->cost);
}
/* }}} mecab_node_toarray */

/* {{{ proto string mecab_node_tostring(resource mecab_node node) */
/**
 * string mecab_node_tostring(resource mecab_node node)
 * string MeCab_Node::toString(void)
 * string MeCab_Node::__toString(void)
 *
 * Get the formatted string of the node.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	string	The formatted string.
 * @see	mecab_format_node
 */
PHP_FUNCTION(mecab_node_tostring)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	/* local variables */
	mecab_t *mecab = NULL;
	const char *fmt = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_NODE_FROM_PARAMETER();

	/* call mecab_format_node() */
	mecab = xnode->tagger->ptr;
	fmt = mecab_format_node(mecab, node);
	if (fmt == NULL) {
		php_mecab_error(E_WARNING, "%s", mecab_strerror(mecab));
		RETURN_FALSE;
	}

	/* set return value */
	RETURN_STRING((char *)fmt, 1);
}
/* }}} mecab_node_tostring */

/* {{{ proto resource mecab_node mecab_node_prev(resource mecab_node node) */
/**
 * resource mecab_node mecab_node_prev(resource mecab_node node)
 * object MeCab_Node MeCab_Node::getPrev(void)
 *
 * Get the previous node.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_node	The previous node.
 *								If the given node is the first one, returns FALSE.
 */
PHP_FUNCTION(mecab_node_prev)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_sibling_wrapper(NODE_PREV);
}
/* }}} mecab_node_prev */

/* {{{ proto resource mecab_node mecab_node_next(resource mecab_node node) */
/**
 * resource mecab_node mecab_node_next(resource mecab_node node)
 * object MeCab_Node MeCab_Node::getNext(void)
 *
 * Get the next node.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_node	The next node.
 *								If the given node is the last one, returns FALSE.
 */
PHP_FUNCTION(mecab_node_next)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_sibling_wrapper(NODE_NEXT);
}
/* }}} mecab_node_next */

/* {{{ proto resource mecab_node mecab_node_enext(resource mecab_node node) */
/**
 * resource mecab_node mecab_node_enext(resource mecab_node node)
 * object MeCab_Node MeCab_Node::getENext(void)
 *
 * Get the enext node.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_node	The next node which has same end point as the given node.
 *								If there is no `enext' node, returns false.
 */
PHP_FUNCTION(mecab_node_enext)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_sibling_wrapper(NODE_ENEXT);
}
/* }}} mecab_node_enext */

/* {{{ proto resource mecab_node mecab_node_bnext(resource mecab_node node) */
/**
 * resource mecab_node mecab_node_bnext(resource mecab_node node)
 * object MeCab_Node MeCab_Node::getBNext(void)
 *
 * Get the bnext node.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_node	The next node which has same beggining point as the given one.
 *								If there is no `bnext' node, returns false.
 */
PHP_FUNCTION(mecab_node_bnext)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_sibling_wrapper(NODE_BNEXT);
}
/* }}} mecab_node_bnext */

/* {{{ proto resource mecab_path mecab_node_rpath(resource mecab_node node) */
/**
 * resource mecab_path mecab_node_rpath(resource mecab_node node)
 * object MeCab_Path MeCab_Node::getRPath(void)
 *
 * Get the rpath.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_path	The next node which has same end point as the given node.
 *								If there is no `rpath' node, returns false.
 */
PHP_FUNCTION(mecab_node_rpath)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_path_wrapper(NODE_RPATH);
}
/* }}} mecab_node_rpath */

/* {{{ proto resource mecab_path mecab_node_lpath(resource mecab_node node) */
/**
 * resource mecab_path mecab_node_lpath(resource mecab_node node)
 * object MeCab_Path MeCab_Node::getLPath(void)
 *
 * Get the lpath.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	resource mecab_path	The next node which has same beggining point as the given one.
 *								If there is no `lpath' node, returns false.
 */
PHP_FUNCTION(mecab_node_lpath)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_get_path_wrapper(NODE_LPATH);
}
/* }}} mecab_node_lpath */

/* {{{ proto resource mecab_node mecab_node_begin_node_list(resource mecab_node node, int index) */
/**
 * resource mecab_node mecab_node_begin_node_list(resource mecab_node node, int index)
 *
 * Get the given index of begin-node list.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @param	int	$index	An indexo of the node.
 * @return	resource mecab_node
 * @see	 mecab_node_end_node_list
 */
PHP_FUNCTION(mecab_node_begin_node_list)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_list_func(0);
}
/* }}} mecab_node_begin_node_list */

/* {{{ proto resource mecab_node mecab_node_end_node_list(resource mecab_node node, int index) */
/**
 * resource mecab_node mecab_node_end_node_list(resource mecab_node node, int index)
 *
 * Get the given index of end-node list.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @param	int	$index	An indexo of the node.
 * @return	resource mecab_node
 * @see	 mecab_node_begin_node_list
 */
PHP_FUNCTION(mecab_node_end_node_list)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_list_func(1);
}
/* }}} mecab_node_end_node_list */

/* {{{ proto string mecab_node_surface(resource mecab_node node) */
/**
 * string mecab_node_surface(resource mecab_node node)
 * string MeCab_Node::getSurface(void)
 *
 * Get the surface.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	string	The surface of the node.
 */
PHP_FUNCTION(mecab_node_surface)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(STRINGL, (char *)node->surface, (int)node->length, 1);
}
/* }}} mecab_node_surface */

/* {{{ proto string mecab_node_feature(resource mecab_node node) */
/**
 * string mecab_node_feature(resource mecab_node node)
 * string MeCab_Node::getFeature(void)
 *
 * Get the feature.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	string	The feature of the node.
 */
PHP_FUNCTION(mecab_node_feature)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(STRING, (char *)node->feature, 1);
}
/* }}} mecab_node_feature */

/* {{{ proto int mecab_node_id(resource mecab_node node) */
/**
 * int mecab_node_id(resource mecab_node node)
 * int MeCab_Node::getId(void)
 *
 * Get the ID.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The ID of the node.
 */
PHP_FUNCTION(mecab_node_id)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->id);
}
/* }}} mecab_node_id */

/* {{{ proto int mecab_node_length(resource mecab_node node) */
/**
 * int mecab_node_length(resource mecab_node node)
 * int MeCab_Node::getLength(void)
 *
 * Get the length of the surface.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The length of the surface of the node.
 */
PHP_FUNCTION(mecab_node_length)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->length);
}
/* }}} mecab_node_length */

/* {{{ proto int mecab_node_rlength(resource mecab_node node) */
/**
 * int mecab_node_rlength(resource mecab_node node)
 * int MeCab_Node::getRLength(void)
 *
 * Get the length of the surface and its leading whitespace.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The length of the surface and its leading whitespace of the node.
 */
PHP_FUNCTION(mecab_node_rlength)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->rlength);
}
/* }}} mecab_node_rlength */

/* {{{ proto int mecab_node_rcattr(resource mecab_node node) */
/**
 * int mecab_node_rcattr(resource mecab_node node)
 * int MeCab_Node::getRcAttr(void)
 *
 * Get the ID of the right context.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The ID of the right context.
 */
PHP_FUNCTION(mecab_node_rcattr)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->rcAttr);
}
/* }}} mecab_node_rcattr */

/* {{{ proto int mecab_node_lcattr(resource mecab_node node) */
/**
 * int mecab_node_lcattr(resource mecab_node node)
 * int MeCab_Node::getLcAttr(void)
 *
 * Get the ID of the left context.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The ID of the left context.
 */
PHP_FUNCTION(mecab_node_lcattr)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->lcAttr);
}
/* }}} mecab_node_lcattr */

/* {{{ proto int mecab_node_posid(resource mecab_node node) */
/**
 * int mecab_node_posid(resource mecab_node node)
 * int MeCab_Node::getPosId(void)
 *
 * Get the ID of the Part-of-Speech.
 * (node->posid is not used in MeCab-0.90)
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The ID of the Part-of-Speech.
 *				Currently, always returns 0.
 */
PHP_FUNCTION(mecab_node_posid)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->posid);
}
/* }}} mecab_node_posid */

/* {{{ proto int mecab_node_char_type(resource mecab_node node) */
/**
 * int mecab_node_char_type(resource mecab_node node)
 * int MeCab_Node::getCharType(void)
 *
 * Get the type of the character.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The type of the character.
 */
PHP_FUNCTION(mecab_node_char_type)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->char_type);
}
/* }}} mecab_node_char_type */

/* {{{ proto int mecab_node_stat(resource mecab_node node) */
/**
 * int mecab_node_stat(resource mecab_node node)
 * int MeCab_Node::getStat(void)
 *
 * Get the status.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The status of the node.
 *				The return value is one of the following:
 *					MECAB_NOR_NODE (0:Normal)
 *					MECAB_UNK_NODE (1:Unknown)
 *					MECAB_BOS_NODE (2:Beginning-of-Sentence)
 *					MECAB_EOS_NODE (3:End-of-Sentence)
 */
PHP_FUNCTION(mecab_node_stat)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->stat);
}
/* }}} mecab_node_stat */

/* {{{ proto bool mecab_node_isbest(resource mecab_node node) */
/**
 * bool mecab_node_isbest(resource mecab_node node)
 * bool MeCab_Node::isBest(void)
 *
 * Determine whether the node is the best solution.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	bool	True if the node is the best, otherwise returns false.
 */
PHP_FUNCTION(mecab_node_isbest)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(BOOL, (node->isbest == 1));
}
/* }}} mecab_node_isbest */

/* {{{ proto int mecab_node_sentence_length(resource mecab_node node) */
/**
 * int mecab_node_sentence_length(resource mecab_node node)
 * int MeCab_Node::getSentenceLength(void)
 *
 * Get the length of the sentence.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The length of the sentence if the node is BOS node, otherwise returns false.
 */
PHP_FUNCTION(mecab_node_sentence_length)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	PHP_MECAB_NODE_FROM_PARAMETER();

	if (node->stat == MECAB_BOS_NODE) {
		RETURN_LONG((long)node->sentence_length);
	} else {
		RETURN_NULL();
	}
}
/* }}} mecab_node_sentence_length */

/* {{{ proto double mecab_node_alpha(resource mecab_node node) */
/**
 * double mecab_node_alpha(resource mecab_node node)
 * double MeCab_Node::getAlpha(void)
 *
 * Get the forward log probability.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	double	The forward log probability of the node.
 */
PHP_FUNCTION(mecab_node_alpha)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(DOUBLE, (double)node->alpha);
}
/* }}} mecab_node_alpha */

/* {{{ proto double mecab_node_beta(resource mecab_node node) */
/**
 * double mecab_node_beta(resource mecab_node node)
 * double MeCab_Node::getBeta(void)
 *
 * Get the backward log probability.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	double	The backward log probability of the node.
 */
PHP_FUNCTION(mecab_node_beta)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(DOUBLE, (double)node->beta);
}
/* }}} mecab_node_beta */

/* {{{ proto double mecab_node_prob(resource mecab_node node) */
/**
 * double mecab_node_prob(resource mecab_node node)
 * double MeCab_Node::getProb(void)
 *
 * Get the marginal probability.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	double	The marginal probability of the node.
 */
PHP_FUNCTION(mecab_node_prob)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(DOUBLE, (double)node->prob);
}
/* }}} mecab_node_prob */

/* {{{ proto int mecab_node_wcost(resource mecab_node node) */
/**
 * int mecab_node_wcost(resource mecab_node node)
 * int MeCab_Node::getWCost(void)
 *
 * Get the word arising cost.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The word arising cost of the node.
 */
PHP_FUNCTION(mecab_node_wcost)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->wcost);
}
/* }}} mecab_node_wcost */

/* {{{ proto int mecab_node_cost(resource mecab_node node) */
/**
 * int mecab_node_cost(resource mecab_node node)
 * int MeCab_Node::getCost(void)
 *
 * Get the cumulative cost.
 *
 * @param	resource mecab_node	$node	The node of the source string.
 * @return	int	The cumulative cost of the node.
 */
PHP_FUNCTION(mecab_node_cost)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_NODE_RETURN_PROPERTY(LONG, (long)node->cost);
}
/* }}} mecab_node_cost */

/* {{{ proto resource mecab_path mecab_path_rnext(resource mecab_path path) */
/**
 * resource mecab_path mecab_path_rnext(resource mecab_path path)
 * object MeCab_Path MeCab_Path::getRNext(void)
 *
 * Get the rnext path.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	resource mecab_path	The rnext path.
 *								If the given path is the first one, returns FALSE.
 */
PHP_FUNCTION(mecab_path_rnext)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_path_get_sibling_wrapper(PATH_RNEXT);
}
/* }}} mecab_path_rnext */

/* {{{ proto resource mecab_path mecab_path_lnext(resource mecab_path path) */
/**
 * resource mecab_path mecab_path_lnext(resource mecab_path path)
 * object MeCab_Path MeCab_Path::getLNext(void)
 *
 * Get the lnext path.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	resource mecab_path	The lnext path.
 *								If the given path is the last one, returns FALSE.
 */
PHP_FUNCTION(mecab_path_lnext)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_path_get_sibling_wrapper(PATH_LNEXT);
}
/* }}} mecab_path_lnext */

/* {{{ proto resource mecab_node mecab_path_rnode(resource mecab_path path) */
/**
 * resource mecab_node mecab_path_rnode(resource mecab_path path)
 * object MeCab_Node MeCab_Path::getRNode(void)
 *
 * Get the rnode.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	resource mecab_node	The next path which has same end point as the given path.
 *								If there is no `rnode' path, returns false.
 */
PHP_FUNCTION(mecab_path_rnode)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_path_get_node_wrapper(PATH_RNODE);
}
/* }}} mecab_path_rnode */

/* {{{ proto resource mecab_node mecab_path_lnode(resource mecab_path path) */
/**
 * resource mecab_node mecab_path_lnode(resource mecab_path path)
 * object MeCab_Node MeCab_Path::getLNode(void)
 *
 * Get the lnode.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	resource mecab_node	The next path which has same beggining point as the given one.
 *								If there is no `lnode' path, returns false.
 */
PHP_FUNCTION(mecab_path_lnode)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_path_get_node_wrapper(PATH_LNODE);
}
/* }}} mecab_path_lnode */

/* {{{ proto double mecab_path_prob(resource mecab_path path) */
/**
 * double mecab_path_prob(resource mecab_path path)
 * double MeCab_Path::getProb(void)
 *
 * Get the marginal probability.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	double	The marginal probability of the path.
 */
PHP_FUNCTION(mecab_path_prob)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_PATH_RETURN_PROPERTY(DOUBLE, (double)(path->prob));
}
/* }}} mecab_path_prob */

/* {{{ proto int mecab_path_cost(resource mecab_path path) */
/**
 * int mecab_path_cost(resource mecab_path path)
 * int MeCab_Path::getCost(void)
 *
 * Get the cumulative cost.
 *
 * @param	resource mecab_path	$path	The path of the source string.
 * @return	int	The cumulative cost of the path.
 */
PHP_FUNCTION(mecab_path_cost)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	PHP_MECAB_PATH_RETURN_PROPERTY(LONG, (long)(path->cost));
}
/* }}} mecab_node_cost */

/* }}} Functions */

#ifdef ZEND_ENGINE_2
/* {{{ methods */

/* {{{ methods of class MeCab_Node*/

/* {{{ proto object MeCab_Node __construct(void) */
/**
 * object MeCab_Node MeCab_Node::__construct(void)
 *
 * Create MeCab_Node object.
 *
 * @access	private
 * @igore
 */
PHP_METHOD(MeCab_Node, __construct)
{
	return;
}
/* }}} MeCab_Node::__construct */

/* {{{ proto mixed MeCab_Node::__get(string name) */
/**
 * mixed MeCab_Node::__get(string name)
 *
 * [Overloading implementation]
 * A magick getter.
 *
 * @param	string	$name	The name of property.
 * @return	mixed	The value of the property.
 *					If there is not a named property, causes E_NOTICE error and returns false.
 * @access	public
 * @ignore
 */
PHP_METHOD(MeCab_Node, __get)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	/* declaration of the arguments */
	char *name = NULL;
	int name_len = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
			&name, &name_len) == FAILURE)
	{
		return;
	} else {
		php_mecab_node_object *intern;
		PHP_MECAB_FETCH_OBJECT(intern, php_mecab_node_object *, object);
		xnode = intern->ptr;
		node = xnode->ptr;
	}

	/* check for given property name */
	if (!strcmp(name, "prev")) {
		php_mecab_node_get_sibling(return_value, object, xnode, NODE_PREV);
		return;
	}
	if (!strcmp(name, "next")) {
		php_mecab_node_get_sibling(return_value, object, xnode, NODE_NEXT);
		return;
	}
	if (!strcmp(name, "enext")) {
		php_mecab_node_get_sibling(return_value, object, xnode, NODE_ENEXT);
		return;
	}
	if (!strcmp(name, "bnext")) {
		php_mecab_node_get_sibling(return_value, object, xnode, NODE_BNEXT);
		return;
	}
	if (!strcmp(name, "rpath")) {
		php_mecab_node_get_path(return_value, object, xnode, NODE_RPATH);
		return;
	}
	if (!strcmp(name, "lpath")) {
		php_mecab_node_get_path(return_value, object, xnode, NODE_LPATH);
		return;
	}
	if (!strcmp(name, "surface"))   RETURN_STRINGL((char *)node->surface, (int)node->length, 1);
	if (!strcmp(name, "feature"))   RETURN_STRING((char *)node->feature, 1);
	if (!strcmp(name, "id"))        RETURN_LONG((long)node->id);
	if (!strcmp(name, "length"))    RETURN_LONG((long)node->length);
	if (!strcmp(name, "rlength"))   RETURN_LONG((long)node->rlength);
	if (!strcmp(name, "rcAttr"))    RETURN_LONG((long)node->rcAttr);
	if (!strcmp(name, "lcAttr"))    RETURN_LONG((long)node->lcAttr);
	if (!strcmp(name, "posid"))     RETURN_LONG((long)node->posid);
	if (!strcmp(name, "char_type")) RETURN_LONG((long)node->char_type);
	if (!strcmp(name, "stat"))      RETURN_LONG((long)node->stat);
	if (!strcmp(name, "isbest"))    RETURN_BOOL((long)node->isbest);
	if (!strcmp(name, "sentence_length")) {
		if (node->stat == MECAB_BOS_NODE) {
			RETURN_LONG((long)node->sentence_length);
		} else {
			RETURN_NULL();
		}
	}
	if (!strcmp(name, "alpha"))     RETURN_DOUBLE((double)node->alpha);
	if (!strcmp(name, "beta"))      RETURN_DOUBLE((double)node->beta);
	if (!strcmp(name, "prob"))      RETURN_DOUBLE((double)node->prob);
	if (!strcmp(name, "wcost"))     RETURN_LONG((long)node->wcost);
	if (!strcmp(name, "cost"))      RETURN_LONG((long)node->cost);

	/* when going to fetch undefined property */
	php_mecab_error(E_NOTICE, "Undefined property (%s)", name);
	RETURN_NULL();
}
/* }}} MeCab_Node::__get */

/* {{{ proto bool MeCab_Node::__isset(string name) */
/**
 * bool MeCab_Node::__isset(string name)
 *
 * [Overloading implementation]
 * Determine whether there is a named property.
 *
 * @param	string	$name	The name of property.
 * @return	bool	True if there is a named property, otherwise returns false.
 * @access	public
 * @ignore
 */
PHP_METHOD(MeCab_Node, __isset)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *znode = NULL;
	php_mecab_node *xnode = NULL;
	const mecab_node_t *node = NULL;

	/* declaration of the arguments */
	char *name = NULL;
	int name_len = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
			&name, &name_len) == FAILURE)
	{
		return;
	} else {
		php_mecab_node_object *intern;
		PHP_MECAB_FETCH_OBJECT(intern, php_mecab_node_object *, object);
		xnode = intern->ptr;
		node = xnode->ptr;
	}

	/* check for given property name */
	if ((!strcmp(name, "prev") && node->prev != NULL) ||
		(!strcmp(name, "next") && node->next != NULL) ||
		(!strcmp(name, "enext") && node->enext != NULL) ||
		(!strcmp(name, "bnext") && node->bnext != NULL) ||
		(!strcmp(name, "rpath") && node->rpath != NULL) ||
		(!strcmp(name, "lpath") && node->lpath != NULL) ||
		!strcmp(name, "surface") ||
		!strcmp(name, "feature") ||
		!strcmp(name, "id") ||
		!strcmp(name, "length") ||
		!strcmp(name, "rlength") ||
		!strcmp(name, "rcAttr") ||
		!strcmp(name, "lcAttr") ||
		!strcmp(name, "posid") ||
		!strcmp(name, "char_type") ||
		!strcmp(name, "stat") ||
		!strcmp(name, "isbest") ||
		!strcmp(name, "sentence_length") ||
		!strcmp(name, "alpha") ||
		!strcmp(name, "beta") ||
		!strcmp(name, "prob") ||
		!strcmp(name, "wcost") ||
		!strcmp(name, "cost"))
	{
		RETURN_TRUE;
	}
	RETURN_FALSE;
}
/* }}} MeCab_Node::__isset */

/* {{{ proto object MeCab_NodeIterator MeCab_Node::getIterator(void) */
/**
 * object MeCab_NodeIterator MeCab_Node::getIterator(void)
 *
 * [IteratorAggregate implementation]
 * Return the iterator object.
 *
 * @return	object MeCab_NodeIterator
 * @access	public
 * @ignore
 */
PHP_METHOD(MeCab_Node, getIterator)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;
	const mecab_node_t *node;
	php_mecab_node_object *newobj;
	php_mecab_node *newnode;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	if (node == NULL) {
		RETURN_NULL();
	}

	php_mecab_object_init(return_value, mecab_iterator_ce);
	PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, return_value);
	newobj->root = node;
	newobj->mode = intern->mode;
	newnode = newobj->ptr;
	newnode->ptr = node;
	php_mecab_node_set_tagger(newnode, xnode->tagger);
}
/* }}} MeCab_Node::getIterator */

/* {{{ proto void MeCab_Node::setTraverse(int tranverse) */
/**
 * void MeCab_Node::setTraverse(int tranverse)
 *
 * Set the traverse mode.
 *
 * @param	int	$traverse	The traverse mode.
 * @return	void
 * @throws	InvalidArgumentException
 * @access	public
 */
PHP_METHOD(MeCab_Node, setTraverse)
{
	php_mecab_node_object *intern;
	long traverse = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_set_error_handling(EH_THROW, spl_ce_InvalidArgumentException TSRMLS_CC);
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l",
			&traverse) == FAILURE)
	{
		php_std_error_handling();
		return;
	}
	php_std_error_handling();

	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);

	if (traverse == (long)TRAVERSE_NEXT ||
		traverse == (long)TRAVERSE_ENEXT ||
		traverse == (long)TRAVERSE_BNEXT)
	{
		intern->mode = (php_mecab_traverse_mode)traverse;
	} else {
		zend_throw_exception_ex(spl_ce_InvalidArgumentException, 0 TSRMLS_CC,
				"Invalid traverse mdoe given");
	}
}
/* }}} MeCab_Node::setTraverse */

/* {{{ proto object MeCab_Node MeCab_Node::getBeginNodeList(int index) */
/**
 * object MeCab_Node MeCab_Node::getBeginNodeList(int index)
 *
 * Return the iterator object for the given index of begin-node list.
 *
 * @param	int	$index	An indexo of the node.
 * @return	object MeCab_Node
 * @throws	BadMethodCallException, InvalidArgumentException, OutOfBoundsException
 * @access	public
 * @see	 MeCab_Node::getEndNodeList
 */
PHP_METHOD(MeCab_Node, getBeginNodeList)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_list_method(0);
}
/* }}} MeCab_Node::getBeginNodeList */

/* {{{ proto object MeCab_Node MeCab_Node::getEndNodeList(int index) */
/**
 * object MeCab_Node MeCab_Node::getEndNodeList(int index)
 *
 * Return the node object for the given index of end-node list.
 *
 * @param	int	$index	An indexo of the node.
 * @return	object MeCab_Node
 * @throws	BadMethodCallException, InvalidArgumentException, OutOfBoundsException
 * @access	public
 * @see	 MeCab_Node::getBeginNodeList
 */
PHP_METHOD(MeCab_Node, getEndNodeList)
{
	__debug_print_current_func(__FUNCTION__, __LINE__);
	php_mecab_node_list_method(1);
}
/* }}} MeCab_Node::getEndNodeList */

/* }}} methods of class MeCab_Node */

/* {{{ methods of class MeCab_NodeIterator*/

/* {{{ proto object MeCab_NodeIterator __construct(void) */
/**
 * object MeCab_Path MeCab_NodeIterator::__construct(void)
 *
 * Create MeCab_NodeIterator object.
 *
 * @access	private
 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, __construct)
{
	return;
}
/* }}} MeCab_NodeIterator::__construct */

/* {{{ object MeCab_Node MeCab_NodeIterator::current(void) */
/**
 * object MeCab_Node MeCab_NodeIterator::current(void)
 *
 * [Iterator implementation]
 * Return the current element.
 *
 * @access	public
 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, current)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;
	const mecab_node_t *node;
	php_mecab_node_object *newobj;
	php_mecab_node *newnode;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	if (node == NULL) {
		RETURN_NULL();
	}

	php_mecab_object_init(return_value, mecab_node_ce);
	PHP_MECAB_FETCH_OBJECT(newobj, php_mecab_node_object *, return_value);
	newobj->mode = intern->mode;
	newnode = newobj->ptr;
	newnode->ptr = node;
	php_mecab_node_set_tagger(newnode, xnode->tagger);
}
/* }}} MeCab_NodeIterator::current */

/* {{{ proto int MeCab_NodeIterator::key(void) */
/**
 * int MeCab_Node::key(void)
 *
 * [Iterator implementation]
 * Return the key of the current element.
 *
 * @return	int	The cumulative cost of the node.
 * @access	public
 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, key)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;
	const mecab_node_t *node;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	if (node == NULL) {
		RETURN_NULL();
	}

	RETURN_LONG((long)node->id);
}
/* }}} MeCab_NodeIterator::key */

/* {{{ proto void MeCab_NodeIterator::next(void) */
/**
 * void MeCab_NodeIterator::next(void)
 *
 * [Iterator implementation]
 * Set the node pointer to the next.
 *
 * @return	void
 * @access	public
 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, next)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;
	const mecab_node_t *node;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	if (node == NULL) {
		return;
	}

	switch (intern->mode) {
	  case TRAVERSE_NEXT:
		xnode->ptr = node->next;
		break;
	  case TRAVERSE_ENEXT:
		xnode->ptr = node->enext;
		break;
	  case TRAVERSE_BNEXT:
		xnode->ptr = node->bnext;
		break;
	  default:
		xnode->ptr = NULL;
	}
}
/* }}} MeCab_NodeIterator::next */

/* {{{ proto void MeCab_NodeIterator::rewind(void) */
/**
 * void MeCab_NodeIterator::rewind(void)
 *
 * [Iterator implementation]
 * Set the node pointer to the beginning.
 *
 * @return	void
 * @access	public

 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, rewind)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	xnode->ptr = intern->root;
}
/* }}} MeCab_NodeIterator::rewind */

/* {{{ proto bool MeCab_NodeIterator::valid(void) */
/**
 * bool MeCab_NodeIterator::valid(void)
 *
 * [Iterator implementation]
 * Check if there is a current element after calls to rewind() or next().
 *
 * @return	bool	True if there is an element after the current element, otherwise returns false.
 * @access	public
 * @igore
 */
PHP_METHOD(MeCab_NodeIterator, valid)
{
	php_mecab_node_object *intern;
	php_mecab_node *xnode;
	const mecab_node_t *node;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	if (ZEND_NUM_ARGS() != 0) {
		WRONG_PARAM_COUNT;
	}
	intern = (php_mecab_node_object *)zend_object_store_get_object(getThis() TSRMLS_CC);
	xnode = intern->ptr;
	node = xnode->ptr;

	RETURN_BOOL(node != NULL);
}
/* }}} MeCab_NodeIterator::valid */

/* }}} methods of class MeCab_NodeIterator */

/* {{{ methods of class MeCab_Path*/

/* {{{ proto object MeCab_Path __construct(void) */
/**
 * object MeCab_Path MeCab_Path::__construct(void)
 *
 * Create MeCab_Path object.
 *
 * @access	private
 * @igore
 */
PHP_METHOD(MeCab_Path, __construct)
{
	return;
}
/* }}} MeCab_Path::__construct */

/* {{{ proto mixed MeCab_Path::__get(string name) */
/**
 * mixed MeCab_Path::__get(string name)
 *
 * [Overloading implementation]
 * A magick getter.
 *
 * @param	string	$name	The name of property.
 * @return	mixed	The value of the property.
 *					If there is not a named property, causes E_NOTICE error and returns false.
 * @access	public
 * @ignore
 */
PHP_METHOD(MeCab_Path, __get)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zpath = NULL;
	php_mecab_path *xpath = NULL;
	const mecab_path_t *path = NULL;

	/* declaration of the arguments */
	char *name = NULL;
	int name_len = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
			&name, &name_len) == FAILURE)
	{
		return;
	} else {
		php_mecab_path_object *intern;
		PHP_MECAB_FETCH_OBJECT(intern, php_mecab_path_object *, object);
		xpath = intern->ptr;
		path = xpath->ptr;
	}

	/* check for given property name */
	if (!strcmp(name, "rnext")) {
		php_mecab_path_get_sibling(return_value, object, xpath, PATH_RNEXT);
		return;
	}
	if (!strcmp(name, "lnext")) {
		php_mecab_path_get_sibling(return_value, object, xpath, PATH_LNEXT);
		return;
	}
	if (!strcmp(name, "rnode")) {
		php_mecab_path_get_node(return_value, object, xpath, PATH_RNODE);
		return;
	}
	if (!strcmp(name, "lnode")) {
		php_mecab_path_get_node(return_value, object, xpath, PATH_LNODE);
		return;
	}
	if (!strcmp(name, "prob")) RETURN_DOUBLE((double)(path->prob));
	if (!strcmp(name, "cost")) RETURN_LONG((long)(path->cost));

	/* when going to fetch undefined property */
	php_mecab_error(E_NOTICE, "Undefined property (%s)", name);
	RETURN_NULL();
}
/* }}} MeCab_Path::__get */

/* {{{ proto boolMeCab_Path:: __isset(string name) */
/**
 * bool MeCab_Path::__isset(string name)
 *
 * [Overloading implementation]
 * Determine whether there is a named property.
 *
 * @param	string	$name	The name of property.
 * @return	bool	True if there is a named property, otherwise returns false.
 * @access	public
 * @ignore
 */
PHP_METHOD(MeCab_Path, __isset)
{
	/* declaration of the resources */
	zval *object = getThis();
	zval *zpath = NULL;
	php_mecab_path *xpath = NULL;
	const mecab_path_t *path = NULL;

	/* declaration of the arguments */
	char *name = NULL;
	int name_len = 0;

	__debug_print_current_func(__FUNCTION__, __LINE__);
	/* parse the arguments */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &name, &name_len) == FAILURE) {
		return;
	} else {
		php_mecab_path_object *intern;
		PHP_MECAB_FETCH_OBJECT(intern, php_mecab_path_object *, object);
		xpath = intern->ptr;
		path = xpath->ptr;
	}

	/* check for given property name */
	if ((!strcmp(name, "rnext") && path->rnext != NULL) ||
		(!strcmp(name, "lnext") && path->lnext != NULL) ||
		(!strcmp(name, "rnode") && path->rnode != NULL) ||
		(!strcmp(name, "lnode") && path->lnode != NULL) ||
		!strcmp(name, "prob") ||
		!strcmp(name, "cost"))
	{
		RETURN_TRUE;
	}
	RETURN_FALSE;
}
/* }}} MeCab_Path::__isset */

/* }}} methods of class MeCab_Path */

/* }}} methods */
#endif /* ZEND_ENGINE_2 */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
